from typing import List, Dict, Type

import pandas as pd

from magnumapi.geometry.GeometryBuilder import GeometryBuilder, SlottedGeometryBuilder
from magnumapi.geometry.SlottedGeometry import SlottedGeometry, SlottedRelativeCosThetaGeometry
from magnumapi.geometry.blocks.CosThetaBlock import AbsoluteCosThetaBlock, RelativeCosThetaBlock
from magnumapi.geometry.Geometry import Geometry, RelativeCosThetaGeometry
from magnumapi.geometry.blocks.RectangularBlock import RectangularBlock
from magnumapi.geometry.definitions.LayerDefinition import LayerDefinition, SlottedLayerDefinition
from magnumapi.geometry.definitions.AbsoluteCosThetaBlockDefinition import AbsoluteCosThetaBlockDefinition
from magnumapi.geometry.definitions.RectangularBlockDefinition import RectangularBlockDefinition
from magnumapi.geometry.definitions.RelativeCosThetaBlockDefinition import RelativeCosThetaBlockDefinition
from magnumapi.cadata.CableDatabase import CableDatabase
import magnumapi.commons.json_file as json_file
import magnumapi.tool_adapters.roxie.RoxieAPI as RoxieAPI


class GeometryFactory:
    """ GeometryFactory implements a factory design pattern and is used to produce:
    - rectangular geometry
    - absolute cos-theta geometry
    - relative cos-theta geometry
    - slotted absolute cos-theta geometry
    - slotted relative cos-theta geometry

    """

    @classmethod
    def init_with_json(cls, json_file_path: str, cadata: CableDatabase) -> Type[Geometry]:
        """ Class method initializing a Geometry instance from a JSON file.

        :param json_file_path: a path to a json file
        :param cadata: a CableDatabase instance
        :return: initialized geometry instance
        """
        json_content = json_file.read(json_file_path)
        block_dct_defs = json_content['block_defs']
        layer_dct_defs = json_content['layer_defs']
        return cls.init_with_dict(block_dct_defs, layer_dct_defs, cadata)

    @classmethod
    def init_with_dict(cls,
                       block_dct_defs: List[Dict],
                       layer_dct_defs: List[Dict],
                       cadata: CableDatabase) -> Type[Geometry]:
        """ Class method initializing a Geometry instance from a list of dictionaries with block definition.

        :param block_dct_defs: a list of dictionaries with geometry definition (block definition)
        :param layer_dct_defs: a list of dictionaries with layer definitions
        :param cadata: a CableDatabase instance
        :return: initialized geometry instance
        """
        return GeometryBuilder() \
            .with_block_defs(block_dct_defs, cadata) \
            .with_layer_defs(layer_dct_defs) \
            .build()

    @classmethod
    def init_slotted_with_json(cls, json_file_path: str, cadata: CableDatabase) -> Type[Geometry]:
        """ Class method initializing a Geometry instance from a JSON file.

        :param json_file_path: a path to a json file
        :param cadata: a CableDatabase instance
        :return: initialized geometry instance
        """
        json_content = json_file.read(json_file_path)
        block_dct_defs = json_content['block_defs']
        layer_dct_defs = json_content['layer_defs']
        r_aperture = json_content['r_aperture']
        return cls.init_slotted_with_dict(block_dct_defs, layer_dct_defs, cadata, r_aperture=r_aperture)

    @classmethod
    def init_slotted_with_dict(cls,
                               block_dct_defs: List[Dict],
                               layer_dct_defs: List[Dict],
                               cadata: CableDatabase,
                               r_aperture: float) -> Type[Geometry]:
        """ Class method initializing a Geometry instance from a list of dictionaries with block definition.

        :param block_dct_defs: a list of dictionaries with geometry definition (block definition)
        :param layer_dct_defs: a list of dictionaries with layer definitions
        :param cadata: a CableDatabase instance
        :param r_aperture: aperture radius in mm
        :return: initialized geometry instance
        """
        return SlottedGeometryBuilder() \
            .with_block_defs(block_dct_defs, cadata) \
            .with_layer_defs(layer_dct_defs) \
            .with_r_aperture(r_aperture) \
            .build()

    @classmethod
    def init_with_data(cls, data_file_path: str, cadata: CableDatabase) -> Geometry:
        """ Class method initializing a Geometry instance from a DATA ROXIE file.

        :param data_file_path: a path to a json file
        :param cadata: a CableDatabase instance
        :return: initialized geometry instance
        """
        block_df = RoxieAPI.read_bottom_header_table(data_file_path, keyword='BLOCK')
        layer_df = RoxieAPI.read_nested_bottom_header_table(data_file_path, keyword='LAYER')
        return cls.init_with_df(block_df, layer_df, cadata)

    @classmethod
    def init_with_csv(cls,
                      block_csv_file_path: str,
                      layer_csv_file_path: str,
                      cadata: CableDatabase) -> Geometry:
        """ Class method initializing a Geometry instance from a CSV file.

        :param block_csv_file_path: a path to a csv file with block definitions
        :param layer_csv_file_path: a path to a csv file with layer definitions
        :param cadata: a CableDatabase instance
        :return: initialized geometry instance
        """
        block_df = pd.read_csv(block_csv_file_path, index_col=0)
        layer_df = pd.read_csv(layer_csv_file_path, index_col=0)
        return cls.init_with_df(block_df, layer_df, cadata)

    @classmethod
    def init_with_df(cls, block_df: pd.DataFrame, layer_df, cadata: CableDatabase) -> Geometry:
        """ Class method initializing a Geometry instance from a dataframe with block definition.

        :param block_df: a dataframe with geometry definition (block definition)
        :param cadata: a CableDatabase instance
        :return: initialized geometry instance
        """
        return GeometryBuilder() \
            .with_block_df(block_df, cadata) \
            .with_layer_df(layer_df) \
            .build()

    @classmethod
    def init_slotted_with_data(cls, data_file_path: str, cadata: CableDatabase, r_aperture: float) -> Geometry:
        """ Class method initializing a Geometry instance from a DATA ROXIE file.

        :param data_file_path: a path to a json file
        :param cadata: a CableDatabase instance
        :param r_aperture: aperture radius in mm
        :return: initialized geometry instance
        """
        block_df = RoxieAPI.read_bottom_header_table(data_file_path, keyword='BLOCK')
        layer_df = RoxieAPI.read_nested_bottom_header_table(data_file_path, keyword='LAYER')
        return cls.init_slotted_with_df(block_df, layer_df, cadata, r_aperture)

    @classmethod
    def init_with_csv_slotted(cls,
                      block_csv_file_path: str,
                      layer_csv_file_path: str,
                      cadata: CableDatabase,
                      r_aperture: float) -> Geometry:
        """ Class method initializing a Geometry instance from a CSV file.

        :param block_csv_file_path: a path to a csv file with block definitions
        :param layer_csv_file_path: a path to a csv file with layer definitions
        :param cadata: a CableDatabase instance
        :param r_aperture: aperture radius in mm
        :return: initialized geometry instance
        """
        block_df = pd.read_csv(block_csv_file_path, index_col=0)
        layer_df = pd.read_csv(layer_csv_file_path, index_col=0)
        return cls.init_slotted_with_df(block_df, layer_df, cadata, r_aperture)

    @classmethod
    def init_slotted_with_df(cls,
                             block_df: pd.DataFrame,
                             layer_df: pd.DataFrame,
                             cadata: CableDatabase,
                             r_aperture: float) -> Geometry:
        """ Class method initializing a Geometry instance from a dataframe with block definition.

        :param block_df: a dataframe with geometry definition (block definition)
        :param cadata: a CableDatabase instance
        :param r_aperture: aperture radius in mm
        :return: initialized geometry instance
        """
        return SlottedGeometryBuilder() \
            .with_block_df(block_df, cadata) \
            .with_layer_df(layer_df) \
            .with_r_aperture(r_aperture) \
            .build()



from typing import List
import os

from magnumapi.geometry.Geometry import HomogenizedCosThetaGeometry
from magnumapi.geometry.SlottedGeometry import SlottedHomogenizedCosThetaGeometry
import magnumapi.commons.text_file as text_file


class AnsysInputBuilder:
    """ An AnsysInputBuilder class with methods used to create an ANSYS coil geometry input

    """

    @classmethod
    def update_input_template(cls, template_path: str, index: int, input_dir: str, input_file: str):
        # read template file
        template_content = text_file.read(template_path)

        # update template file with index
        template_content = template_content.format(index)

        # write updated template file to output directory with index
        output_path = os.path.join(input_dir, input_file % index)
        template_content = text_file.write(output_path, template_content)

    @classmethod
    def write_geometry_input_file(cls, homo_geometry: HomogenizedCosThetaGeometry, model_path: str) -> None:
        """ Class method writing a geometry parameters for ANSYS input file from a homogenized cos-theta geometry

        :param homo_geometry: a homogenized cos-theta geometry
        :param model_path: a path to model input
        """
        output_text = cls.prepare_ansys_model_input(homo_geometry)
        text_file.writelines(model_path, output_text)

    @staticmethod
    def prepare_ansys_model_input(homo_geometry: HomogenizedCosThetaGeometry) -> List[str]:
        """ Static method preparing an ANSYS model input

        :param homo_geometry:
        :return:
        """
        n_layers = homo_geometry.get_number_of_layers()
        blocks_per_layer = homo_geometry.get_number_of_blocks_per_layer()
        return AnsysInputBuilder.generate_ansys_input_text(homo_geometry, n_layers, blocks_per_layer)

    @staticmethod
    def generate_ansys_input_text(homo_geometry: HomogenizedCosThetaGeometry, n_layers, blocks_per_layer) -> List[str]:
        """ Method generating an ANSYS coil geometry input text from a homogenized cos-theta geometry

        :param homo_geometry: homogenized cos-theta geometry
        :param n_layers: number of layers
        :param blocks_per_layer: number of blocks per layer
        :return: a list of model inputs
        """
        model_inputs = []
        model_inputs.append('! Example Input File\n')
        model_inputs.append('! -------------------\n')
        model_inputs.append('! - Standard\n')
        model_inputs.append('/UNITS,SI\n')
        model_inputs.append('pi = acos(-1)\n')
        model_inputs.append('*afun,deg		 ! Specifies the units for angular functions in parameter expressions\n')
        model_inputs.append('tref,293\n')
        model_inputs.append('! -------------------\n')
        model_inputs.append('! Set-up arrays\n')
        model_inputs.append('! *dim,V_Field,array,1,1\n')
        model_inputs.append('! *dim,V_Nlay,array,1,1\n')
        model_inputs.append('! *dim,V_blocks,array,1,1\n')
        model_inputs.append('! *dim,V_current,array,1,1\n')
        model_inputs.append('! *dim,V_rad,array,1,2\n')
        model_inputs.append('! *dim,V_turns,array,1,1\n')
        model_inputs.append('! *dim,V_corners,array,2,4\n')
        model_inputs.append('\n')
        model_inputs.append('! General\n')

        model_inputs.append('Nlay = %d\n' % n_layers)
        model_inputs.append('\n')

        radius_inner_prev = None
        index_layer = -1
        index_block = 0
        for block in homo_geometry.blocks:
            if radius_inner_prev != block.block_def.radius_inner:
                index_block = 0
                index_layer += 1
                model_inputs.append('! Layer\n')
                model_inputs.append('Nb%d = %d\n' % (index_layer + 1, blocks_per_layer[index_layer]))
                model_inputs.append('current%d = %f\n' % (index_layer + 1, block.block_def.current))
                model_inputs.append('r%d1 = %fe-3\n' % (index_layer + 1, block.block_def.radius_inner))
                model_inputs.append('r%d2 = %fe-3\n' % (index_layer + 1, block.block_def.radius_outer))

            model_inputs.append('! Block\n')
            model_inputs.append('Nc%d%d = %d\n' % (index_layer + 1, index_block + 1, block.block_def.nco))
            model_inputs.append('\n')
            model_inputs.append('! Save angles\n')
            model_inputs.append('t%d%d_1 = %.4f\n' % (index_layer + 1, index_block + 1, block.block_def.phi_0))
            model_inputs.append('t%d%d_2 = %.4f\n' % (index_layer + 1, index_block + 1, block.block_def.phi_1))
            model_inputs.append('t%d%d_3 = %.4f\n' % (index_layer + 1, index_block + 1, block.block_def.phi_2))
            model_inputs.append('t%d%d_4 = %.4f\n' % (index_layer + 1, index_block + 1, block.block_def.phi_3))

            if index_block + 1 == blocks_per_layer[index_layer]:
                # Fixes the first 2 angles of the 1st block to be 0 / assumes mid-plane shim thickness = 0
                model_inputs.append('\n')
                model_inputs.append('t%d1_1 = 0\n' % (index_layer + 1))
                model_inputs.append('t%d1_2 = 0\n' % (index_layer + 1))
                model_inputs.append('\n')

            index_block += 1
            radius_inner_prev = block.block_def.radius_inner

        model_inputs.append('! Names\n')
        model_inputs.append('r1 = r11\n')
        model_inputs.append('r%d = r%d2\n' % (n_layers + 1, n_layers))
        model_inputs.append('rout = r%d\n' % (n_layers + 1))
        model_inputs.append('Nlayers = Nlay\n')

        model_inputs.append('\n')
        model_inputs.append('! -------------------------------------\n')
        model_inputs.append('! - Geometry parameters\n')
        model_inputs.append('! ---- Filler\n')
        model_inputs.append('*set, fillerth, 5e-3\n')

        model_inputs.append('\n')
        model_inputs.append('! ---- Yoke\n')
        model_inputs.append('*set,yoketh,260e-3\n')

        model_inputs.append('\n')
        model_inputs.append('! Update geometry\n')
        model_inputs.append('*set, rin_yoke, rout + fillerth\n')
        model_inputs.append('*set, rout_yoke, rin_yoke + yoketh\n')

        model_inputs.append('\n')
        model_inputs.append('! -------------------------------------\n')
        model_inputs.append('! Interlay Contact - Filler is Nlayers+1\n')
        model_inputs.append('! Glue = 5 , Sliding = 0\n')
        model_inputs.append('ilay_12 = 5\n')
        model_inputs.append('ilay_23 = 0\n')
        model_inputs.append('ilay_34 = 5\n')
        model_inputs.append('ilay_45 = 0\n')
        model_inputs.append('! -------------------------------------\n')
        model_inputs.append('! Mesh Parameters\n')
        model_inputs.append('mpar = 1\n')
        model_inputs.append('\n')
        model_inputs.append('mesh_azim_size = mpar*2e-3    ! Coil Azimuthal\n')
        model_inputs.append('mesh_radial_size = mpar*2e-3    ! Coil Radial\n')
        model_inputs.append('\n')
        model_inputs.append('msize_aperture = mpar*1e-3\n')
        model_inputs.append('msize_filler = mpar*3e-3\n')
        model_inputs.append('msize_yoke = mpar*10e-3\n')
        model_inputs.append('\n')
        model_inputs.append('!!!!!!!!!!!!!!!!!!!!!!!\n')
        model_inputs.append('! Friction Parameters !\n')
        model_inputs.append('!!!!!!!!!!!!!!!!!!!!!!!\n')
        model_inputs.append('mu_single = 0.0\n')
        model_inputs.append('mu_g10_coil = mu_single\n')
        model_inputs.append('mu_ti_coil = mu_single\n')
        model_inputs.append('mu_ti_g10 = mu_single\n')
        model_inputs.append('mu_g10_ss = mu_single\n')
        model_inputs.append('mu_alu_iron = mu_single\n')
        model_inputs.append('mu_ss_ss = mu_single\n')
        model_inputs.append('mu_ss_ti = mu_single\n')
        model_inputs.append('mu_ss_iron = mu_single\n')

        return model_inputs



import pandas as pd
import pytest

from magnumapi.geometry.GeometryFactory import GeometryFactory
from magnumapi.cadata.CableDatabase import CableDatabase
import magnumapi.commons.json_file as json_file
from tests.resource_files import create_resources_file_path


def test_to_dict():
    # arrange
    json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_rel.json')
    cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
    cadata = CableDatabase.read_cadata(cadata_path)

    # act
    geometry = GeometryFactory.init_with_json(json_path, cadata)
    geometry.build_blocks()
    block_layer_defs = geometry.to_abs_geometry().to_dict()

    # assert
    json_ref_path = create_resources_file_path('resources/geometry/roxie/16T/16T_abs.json')
    block_layer_ref_defs = json_file.read(json_ref_path)

    pd.testing.assert_frame_equal(pd.DataFrame(block_layer_ref_defs['block_defs']).sort_index(axis=1),
                                  pd.DataFrame(block_layer_defs['block_defs']).sort_index(axis=1))

    pd.testing.assert_frame_equal(pd.DataFrame(block_layer_ref_defs['layer_defs']),
                                  pd.DataFrame(block_layer_defs['layer_defs']))


def test_to_abs_geometry():
    # arrange
    json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_rel.json')
    cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
    cadata = CableDatabase.read_cadata(cadata_path)

    # act
    geometry = GeometryFactory.init_with_json(json_path, cadata)
    geometry = geometry.to_abs_geometry()
    block_layer_defs = geometry.to_dict()

    # assert
    json_ref_path = create_resources_file_path('resources/geometry/roxie/16T/16T_abs.json')
    block_layer_ref_defs = json_file.read(json_ref_path)

    pd.testing.assert_frame_equal(pd.DataFrame(block_layer_ref_defs['block_defs']).sort_index(axis=1),
                                  pd.DataFrame(block_layer_defs['block_defs']).sort_index(axis=1))

    pd.testing.assert_frame_equal(pd.DataFrame(block_layer_ref_defs['layer_defs']),
                                  pd.DataFrame(block_layer_defs['layer_defs']))


def test_to_rel_dict():
    # arrange
    json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_abs.json')
    cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
    cadata = CableDatabase.read_cadata(cadata_path)

    # act
    geometry = GeometryFactory.init_with_json(json_path, cadata)
    geometry.build_blocks()
    block_layer_defs = geometry.to_rel_geometry().to_dict()

    # assert
    json_ref_path = create_resources_file_path('resources/geometry/roxie/16T/16T_rel.json')
    block_layer_ref_defs = json_file.read(json_ref_path)

    pd.testing.assert_frame_equal(pd.DataFrame(block_layer_ref_defs['block_defs']).sort_index(axis=1),
                                  pd.DataFrame(block_layer_defs['block_defs']).sort_index(axis=1))

    pd.testing.assert_frame_equal(pd.DataFrame(block_layer_ref_defs['layer_defs']),
                                  pd.DataFrame(block_layer_defs['layer_defs']))


def test_consistency_check_block_numbering_duplicates():
    # arrange
    block_layer_defs = {
        "block_defs": [
            {"no": 1, "radius": 25.0, "alpha": 0, "phi": 0.57294, "nco": 4, "type": 1, "current": 13500,
             "condname": "16TIL9", "n1": 2, "n2": 20, "imag": 0, "turn": 0},
            {"no": 1, "radius": 25.0, "alpha": 26, "phi": 23, "nco": 5, "type": 1, "current": 13500,
             "condname": "16TIL9", "n1": 2, "n2": 20, "imag": 0, "turn": 0}],
        "layer_defs": [{"no": 1, "symm": 1, "typexy": 1, "blocks": [1, 2, 3, 4]}]}
    cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
    cadata = CableDatabase.read_cadata(cadata_path)

    # act
    # assert
    with pytest.raises(AttributeError) as excinfo:
        GeometryFactory.init_with_dict(block_layer_defs['block_defs'],
                                       block_layer_defs['layer_defs'],
                                       cadata)

    assert "The block numbering ([1, 1]) contains duplications!" in str(excinfo.value)


def test_consistency_check_layer_numbering_duplicates():
    # arrange
    block_layer_defs = {
        "block_defs": [
            {"no": 1, "radius": 25.0, "alpha": 0, "phi": 0.57294, "nco": 4, "type": 1, "current": 13500,
             "condname": "16TIL9", "n1": 2, "n2": 20, "imag": 0, "turn": 0},
            {"no": 2, "radius": 25.0, "alpha": 26, "phi": 23, "nco": 5, "type": 1, "current": 13500,
             "condname": "16TIL9", "n1": 2, "n2": 20, "imag": 0, "turn": 0}],
        "layer_defs": [{"no": 1, "symm": 1, "typexy": 1, "blocks": [1, 2]},
                       {"no": 1, "symm": 1, "typexy": 1, "blocks": [1, 2]}]}
    cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
    cadata = CableDatabase.read_cadata(cadata_path)

    # act
    # assert
    with pytest.raises(AttributeError) as excinfo:
        GeometryFactory.init_with_dict(block_layer_defs['block_defs'],
                                       block_layer_defs['layer_defs'],
                                       cadata)

    assert "The layer numbering ([1, 1]) contains duplications!" in str(excinfo.value)


def test_consistency_check_block_layer_numbering_duplicates():
    # arrange
    block_layer_defs = {
        "block_defs": [
            {"no": 1, "radius": 25.0, "alpha": 0, "phi": 0.57294, "nco": 4, "type": 1, "current": 13500,
             "condname": "16TIL9", "n1": 2, "n2": 20, "imag": 0, "turn": 0},
            {"no": 2, "radius": 25.0, "alpha": 26, "phi": 23, "nco": 5, "type": 1, "current": 13500,
             "condname": "16TIL9", "n1": 2, "n2": 20, "imag": 0, "turn": 0}],
        "layer_defs": [{"no": 1, "symm": 1, "typexy": 1, "blocks": [1, 2]},
                       {"no": 2, "symm": 1, "typexy": 1, "blocks": [1, 2]}]}
    cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
    cadata = CableDatabase.read_cadata(cadata_path)

    # act
    # assert
    with pytest.raises(AttributeError) as excinfo:
        GeometryFactory.init_with_dict(block_layer_defs['block_defs'],
                                       block_layer_defs['layer_defs'],
                                       cadata)

    assert "The layer numbering of blocks ([1, 2, 1, 2]) contains duplications!" in str(excinfo.value)


def test_consistency_check_block_and_block_layer_numbering_duplicates():
    # arrange
    block_layer_defs = {
        "block_defs": [
            {"no": 1, "radius": 25.0, "alpha": 0, "phi": 0.57294, "nco": 4, "type": 1, "current": 13500,
             "condname": "16TIL9", "n1": 2, "n2": 20, "imag": 0, "turn": 0},
            {"no": 2, "radius": 25.0, "alpha": 26, "phi": 23, "nco": 5, "type": 1, "current": 13500,
             "condname": "16TIL9", "n1": 2, "n2": 20, "imag": 0, "turn": 0}],
        "layer_defs": [{"no": 1, "symm": 1, "typexy": 1, "blocks": [1, 2]},
                       {"no": 2, "symm": 1, "typexy": 1, "blocks": [3]}]}
    cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
    cadata = CableDatabase.read_cadata(cadata_path)

    # act
    # assert
    with pytest.raises(AttributeError) as excinfo:
        GeometryFactory.init_with_dict(block_layer_defs['block_defs'],
                                       block_layer_defs['layer_defs'],
                                       cadata)

    assert "The numbering in block [1, 2] and layer [1, 2, 3] do not match!" in str(excinfo.value)


def test_to_roxie_layer_df():
    # arrange
    json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_rel.json')
    cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
    cadata = CableDatabase.read_cadata(cadata_path)

    # act
    geometry = GeometryFactory.init_with_json(json_path, cadata)
    geometry.build_blocks()

    df = geometry.to_layer_df()

    # assert
    df_ref = pd.DataFrame({'no': {0: 1, 1: 2, 2: 3, 3: 4},
                           'symm': {0: 1, 1: 1, 2: 1, 3: 1},
                           'typexy': {0: 1, 1: 1, 2: 1, 3: 1},
                           'blocks': {0: '1 2 3 4', 1: '5 6 7', 2: '8 9 10', 3: '11 12'}})
    # print(RoxieAPI.convert_bottom_header_table_to_str(df, keyword='LAYER', suffix=' /'))

    pd.testing.assert_frame_equal(df_ref, df)



import magnumapi.commons.json_file as json_file
import magnumapi.commons.text_file as text_file


def convert_notebook_to_script(input_ipynb_file_path: str, notebook_name: str, output_script_file_path: str) -> None:
    """ Function converting a notebook into a script equivalent. The notebook is executed as a function.
    The function name is given by as run_<notebook_name>. The input arguments are taken from cells with `parameters`
    tags. The return value is taken from the sb.glue code.

    :param input_ipynb_file_path:
    :param notebook_name:
    :param output_script_file_path:
    """
    notebook = json_file.read(input_ipynb_file_path)

    parameters = []
    code_lines = []
    output_line = ''
    for cell in notebook['cells']:
        if cell['cell_type'] == 'code':
            if 'tags' in cell['metadata']:
                if 'parameters' in cell['metadata']['tags']:
                    for line in cell['source']:
                        parameters.append(line)
            else:
                for line in cell['source']:
                    if 'sb.glue' in line:
                        output_line = line
                    else:
                        code_lines.append(line.replace('\n', ''))

    output_var = output_line.split(',')[1]
    if '=' in output_var:
        output_var = output_var.split('=')[-1]

    import_lines = []
    function_lines = []

    for code_line in code_lines:
        if code_line.strip().startswith('from') or code_line.strip().startswith('import'):
            import_lines.append(code_line)
        else:
            if not (code_line.startswith('%') or code_line.startswith('sys.path.append')):
                function_lines.append(code_line)

    output_lines = list(import_lines)

    output_lines.append('\n')

    parameters = [parameter.replace('\n', '') for parameter in parameters]
    output_lines.append('def run_%s_script(%s):' % (notebook_name, ', '.join(parameters)))

    for function_line in function_lines:
        output_lines.append('\t' + function_line)

    output_lines.append('\t' + 'return ' + output_var)

    text_file.writelines(output_script_file_path, output_lines)



import random
from copy import deepcopy
from unittest import TestCase

import pandas as pd
import numpy as np

from magnumapi.geometry.GeometryChange import GeometryChange
from magnumapi.geometry.GeometryFactory import GeometryFactory
from magnumapi.cadata.CableDatabase import CableDatabase
from magnumapi.geometry.SlottedGeometry import SlottedRelativeCosThetaGeometry
from magnumapi.optimization.DesignVariable import GeneticDesignVariable
from magnumapi.optimization.GeneticOptimization import GeneticOptimization, RoxieGeneticOptimization
from tests.resource_files import create_resources_file_path

json_file_path = create_resources_file_path('resources/optimization/config.json')
csv_file_path = create_resources_file_path('resources/optimization/optim_input_enlarged.csv')
optimization_cfg = GeneticOptimization.initialize_config(json_file_path)

json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_rel.json')
cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
cadata = CableDatabase.read_cadata(cadata_path)
geometry = GeometryFactory.init_with_json(json_path, cadata)


class TestRoxieGeneticOptimization(TestCase):

    def setUp(self) -> None:
        self.gen_opt = RoxieGeneticOptimization(config=optimization_cfg,
                                                design_variables_df=pd.read_csv(csv_file_path),
                                                geometry=geometry,
                                                model_input_path='',
                                                is_script_executed=True,
                                                output_subdirectory_dir='')

    def test_initialize_population(self):
        # arrange
        np.random.seed(0)

        # act
        pop = self.gen_opt.initialize_population_old()

        # assert
        pop_ref = [0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 1, 0, 1, 0,
                   1, 1, 0, 1, 1, 0, 0, 1, 0, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1, 0, 1, 0, 0, 1, 1, 0, 1, 0, 1, 0, 0, 0,
                   0, 0, 1, 1, 0, 0, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 0, 1, 0, 0, 1, 1, 0, 1, 0,
                   0, 1, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 1, 0, 1, 1, 0, 1, 1]

        self.assertListEqual(pop_ref, pop[0])

    def test_decode_chromosome(self):
        # arrange
        np.random.seed(0)

        # act
        pop = self.gen_opt.initialize_population_old()
        chromosome = self.gen_opt.decode_individual_old(pop[0])

        # assert
        chromosome_ref = {'phi_r:1:2': 5.953125, 'phi_r:1:3': 9.78125, 'phi_r:1:4': 4.75, 'phi_r:2:2': 5.40625,
                          'phi_r:2:3': 6.28125, 'phi_r:3:2': 7.703125, 'phi_r:3:3': 5.734375, 'phi_r:4:2': 6.390625,
                          'alpha_r:1:2': 3.59375, 'alpha_r:1:3': 6.40625, 'alpha_r:1:4': 6.5625, 'alpha_r:2:2': 0.46875,
                          'alpha_r:2:3': 0.9375, 'alpha_r:3:2': 5.78125, 'alpha_r:3:3': 9.6875, 'alpha_r:4:2': 7.8125,
                          'nco:1:1': 3, 'nco:1:2': 2, 'nco:1:3': 1, 'nco:1:4': 0, 'nco:2:1': 6, 'nco:2:2': 9,
                          'nco:2:3': 0, 'nco:3:1': 14, 'nco:3:2': 9, 'nco:3:3': 3, 'nco:4:1': 30, 'nco:4:2': 12}

        self.assertDictEqual(chromosome_ref, chromosome)

    def test_decode_chromosome_with_global_parameter(self):
        # arrange
        np.random.seed(0)
        csv_global_file_path = create_resources_file_path('resources/optimization/optim_input_enlarged_with_global.csv')
        df = pd.read_csv(csv_global_file_path)
        self.gen_opt.design_variables = RoxieGeneticOptimization.initialize_design_variables(df)

        # act
        pop = self.gen_opt.initialize_population_old()
        chromosome = self.gen_opt.decode_individual_old(pop[0])

        # assert
        chromosome_ref = {'phi_r:1:2': 5.953125, 'phi_r:1:3': 9.78125, 'phi_r:1:4': 4.75, 'phi_r:2:2': 5.40625,
                          'phi_r:2:3': 6.28125, 'phi_r:3:2': 5.734375, 'phi_r:4:2': 6.390625, 'alpha_r:1:2': 3.59375,
                          'alpha_r:1:3': 6.40625, 'alpha_r:1:4': 6.5625, 'alpha_r:2:2': 0.46875, 'alpha_r:2:3': 0.9375,
                          'alpha_r:3:2': 9.6875, 'alpha_r:4:2': 7.8125, 'nco:1:1': 3, 'nco:1:2': 2, 'nco:1:3': 1,
                          'nco:1:4': 0, 'nco:2:1': 6, 'nco:2:2': 9, 'nco:2:3': 0, 'nco:3:1': 14, 'nco:3:2': 9,
                          'nco:3:3': 3, 'nco:4:1': 30, 'nco:4:2': 12, 'R_EE': 1.2109375}

        self.assertDictEqual(chromosome_ref, chromosome)

    def test_decode_chromosome_with_global_parameter_multi_index(self):
        # arrange
        np.random.seed(0)
        path_str = 'resources/optimization/optim_input_enlarged_with_global_multi_index.csv'
        csv_global_file_path = create_resources_file_path(path_str)
        df = pd.read_csv(csv_global_file_path)
        self.gen_opt.design_variables = RoxieGeneticOptimization.initialize_design_variables(df)

        # act
        pop = self.gen_opt.initialize_population_old()
        chromosome = self.gen_opt.decode_individual_old(pop[0])

        # assert
        chromosome_ref = {'phi_r:1:2': 5.953125, 'phi_r:1:3': 9.78125, 'phi_r:1:4': 4.75, 'phi_r:2:2': 5.40625,
                          'phi_r:2:3': 6.28125, 'phi_r:3:2': 7.703125, 'phi_r:3:3': 5.734375, 'phi_r:4:2': 6.390625,
                          'alpha_r:1:2': 3.59375, 'alpha_r:1:3': 6.40625, 'alpha_r:1:4': 6.5625, 'alpha_r:2:2': 0.46875,
                          'alpha_r:2:3': 0.9375, 'alpha_r:3:2': 5.78125, 'alpha_r:3:3': 9.6875, 'alpha_r:4:2': 7.8125,
                          'nco:1:1': 3, 'nco:1:2': 2, 'nco:1:3': 1, 'nco:1:4': 0, 'nco:2:1': 6, 'nco:2:2': 9,
                          'nco:2:3': 0, 'nco:3:1': 14, 'nco:3:2': 9, 'nco:3:3': 3, 'nco:4:1': 30, 'nco:4:2': 12,
                          'R_EE': 1.2109375, 'current:-:1': 12257.8125, 'current:-:2': 12257.8125,
                          'current:-:3': 12257.8125, 'current:-:4': 12257.8125, 'current:-:5': 12257.8125,
                          'current:-:6': 12257.8125, 'current:-:7': 12257.8125, 'current:-:8': 12257.8125,
                          'current:-:9': 12257.8125, 'current:-:10': 12257.8125, 'current:-:11': 12257.8125,
                          'current:-:12': 12257.8125}

        self.assertDictEqual(chromosome_ref, chromosome)

    def test_update_parameters(self):
        # arrange
        np.random.seed(0)

        # act
        pop = self.gen_opt.initialize_population_old()
        chromosome = self.gen_opt.decode_individual_old(pop[0])
        block_layer_defs = self.gen_opt.update_model_parameters(chromosome).to_dict()

        # assert
        block_defs_ref = {'block_defs': [
            {'no': 1, 'radius': 25.0, 'alpha_r': 0, 'phi_r': 0.57294, 'nco': 3, 'type': 1, 'current': 13500,
             'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 2, 'radius': 25.0, 'alpha_r': 3.59375, 'phi_r': 5.953125, 'nco': 2, 'type': 1, 'current': 13500,
             'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 3, 'radius': 25.0, 'alpha_r': 6.40625, 'phi_r': 9.78125, 'nco': 1, 'type': 1, 'current': 13500,
             'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 5, 'radius': 39.0, 'alpha_r': 0.0, 'phi_r': 0.36728, 'nco': 6, 'type': 1, 'current': 13500,
             'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 6, 'radius': 39.0, 'alpha_r': 0.46875, 'phi_r': 5.40625, 'nco': 9, 'type': 1, 'current': 13500,
             'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 8, 'radius': 53.0, 'alpha_r': 0, 'phi_r': 0.27026, 'nco': 14, 'type': 1, 'current': 13500,
             'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 9, 'radius': 53.0, 'alpha_r': 5.78125, 'phi_r': 7.703125, 'nco': 9, 'type': 1, 'current': 13500,
             'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 10, 'radius': 53.0, 'alpha_r': 9.6875, 'phi_r': 5.734375, 'nco': 3, 'type': 1, 'current': 13500,
             'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 11, 'radius': 67.45, 'alpha_r': 0, 'phi_r': 0.21236, 'nco': 30, 'type': 1, 'current': 13500,
             'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 12, 'radius': 67.45, 'alpha_r': 7.8125, 'phi_r': 6.390625, 'nco': 12, 'type': 1, 'current': 13500,
             'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0}],
            'layer_defs': [{'no': 1, 'symm': 1, 'typexy': 1, 'blocks': [1, 2, 3, 4]},
                           {'no': 2, 'symm': 1, 'typexy': 1, 'blocks': [5, 6, 7]},
                           {'no': 3, 'symm': 1, 'typexy': 1, 'blocks': [8, 9, 10]},
                           {'no': 3, 'symm': 1, 'typexy': 1, 'blocks': [11, 12]}]}

        assert block_defs_ref['block_defs'] == block_layer_defs['block_defs']

    def test_update_parameters_multi_index(self):
        # arrange
        np.random.seed(0)
        path_str = 'resources/optimization/optim_input_enlarged_with_global_multi_index.csv'
        csv_global_file_path = create_resources_file_path(path_str)
        df = pd.read_csv(csv_global_file_path)
        self.gen_opt.design_variables = RoxieGeneticOptimization.initialize_design_variables(df)

        # act
        pop = self.gen_opt.initialize_population_old()
        chromosome = self.gen_opt.decode_individual_old(pop[0])
        block_layer_defs = self.gen_opt.update_model_parameters(chromosome).to_dict()

        # assert
        block_defs_ref = [
            {'no': 1, 'radius': 25.0, 'alpha_r': 0, 'phi_r': 0.57294, 'nco': 3, 'type': 1, 'current': 12257.8125,
             'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 2, 'radius': 25.0, 'alpha_r': 3.59375, 'phi_r': 5.953125, 'nco': 2, 'type': 1, 'current': 12257.8125,
             'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 3, 'radius': 25.0, 'alpha_r': 6.40625, 'phi_r': 9.78125, 'nco': 1, 'type': 1, 'current': 12257.8125,
             'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 5, 'radius': 39.0, 'alpha_r': 0.0, 'phi_r': 0.36728, 'nco': 6, 'type': 1, 'current': 12257.8125,
             'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 6, 'radius': 39.0, 'alpha_r': 0.46875, 'phi_r': 5.40625, 'nco': 9, 'type': 1, 'current': 12257.8125,
             'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 8, 'radius': 53.0, 'alpha_r': 0, 'phi_r': 0.27026, 'nco': 14, 'type': 1, 'current': 12257.8125,
             'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 9, 'radius': 53.0, 'alpha_r': 5.78125, 'phi_r': 7.703125, 'nco': 9, 'type': 1, 'current': 12257.8125,
             'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 10, 'radius': 53.0, 'alpha_r': 9.6875, 'phi_r': 5.734375, 'nco': 3, 'type': 1, 'current': 12257.8125,
             'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 11, 'radius': 67.45, 'alpha_r': 0, 'phi_r': 0.21236, 'nco': 30, 'type': 1, 'current': 12257.8125,
             'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 12, 'radius': 67.45, 'alpha_r': 7.8125, 'phi_r': 6.390625, 'nco': 12, 'type': 1,
             'current': 12257.8125, 'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0}]

        self.assertListEqual(block_defs_ref, block_layer_defs['block_defs'])

    def test_correct_missing_blocks_in_block_and_layer_definitions(self):
        # arrange
        chromosome = {'nco_r:1:1': 2,
                      'nco_r:1:2': -3}
        json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_rel.json')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)
        geometry = GeometryFactory.init_with_json(json_path, cadata)

        # act
        # # update number of turns per block_index
        geometry = GeometryChange.update_nco_r(geometry, chromosome)

        assert len(geometry.blocks) == 11
        assert geometry.layer_defs[0].blocks == [1, 3, 4]
        assert geometry.layer_defs[1].blocks == [5, 6, 7]
        assert geometry.layer_defs[2].blocks == [8, 9, 10]
        assert geometry.layer_defs[3].blocks == [11, 12]

    def test_update_targeted_optimization_input_geometry(self):
        np.random.seed(1)

        # act
        path_str = 'resources/optimization/optim_input_enlarged_targeted_optimization.csv'
        csv_global_file_path = create_resources_file_path(path_str)
        df = pd.read_csv(csv_global_file_path)
        self.gen_opt.design_variables = RoxieGeneticOptimization.initialize_design_variables(df)
        pop = self.gen_opt.initialize_population_old()
        chromosome = self.gen_opt.decode_individual_old(pop[0])

        # read an input file with geometry definition
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)
        json_ref_path = create_resources_file_path('resources/geometry/roxie/16T/16T_rel.json')

        # initialize relative geometry
        geometry_rel = GeometryFactory.init_with_json(json_ref_path, cadata)

        # update phi_r
        geometry_rel = GeometryChange.update_phi_r(geometry_rel, chromosome)

        # update nco_r
        geometry_rel = GeometryChange.update_nco_r(geometry_rel, chromosome)

        # extract absolute geometry
        geometry_abs = geometry_rel.to_abs_geometry()

        # correct block radiality
        geometry_abs = GeometryChange.calculate_radial_alpha(geometry_abs)

        # extract relative geometry
        geometry_rel = geometry_abs.to_rel_geometry()

        # update alpha_rad_r with
        geometry_rel = GeometryChange.update_alpha_radial(geometry_rel, chromosome)

        # assert that the number of turns per layer is the same
        assert 13 == sum([block.block_def.nco for block in geometry_rel.blocks[:4]])
        assert 19 == sum([block.block_def.nco for block in geometry_rel.blocks[4:7]])
        assert 29 == sum([block.block_def.nco for block in geometry_rel.blocks[7:10]])
        assert 39 == sum([block.block_def.nco for block in geometry_rel.blocks[10:]])

    def test_update_model_parameters_targeted(self):
        np.random.seed(1)

        # act
        path_str = 'resources/optimization/optim_input_enlarged_targeted_optimization.csv'
        csv_global_file_path = create_resources_file_path(path_str)
        df = pd.read_csv(csv_global_file_path)
        self.gen_opt.design_variables = RoxieGeneticOptimization.initialize_design_variables(df)
        pop = self.gen_opt.initialize_population_old()
        chromosome = self.gen_opt.decode_individual_old(pop[0])
        geometry_rel = self.gen_opt.update_model_parameters_targeted(chromosome)

        # assert that the number of turns per layer is the same
        assert 13 == sum([block.block_def.nco for block in geometry_rel.blocks[:4]])
        assert 19 == sum([block.block_def.nco for block in geometry_rel.blocks[4:7]])
        assert 29 == sum([block.block_def.nco for block in geometry_rel.blocks[7:10]])
        assert 39 == sum([block.block_def.nco for block in geometry_rel.blocks[10:]])

    def test_update_model_parameters_targeted_error(self):
        np.random.seed(1)

        # act
        path_str = 'resources/optimization/optim_input_enlarged_targeted_optimization.csv'
        csv_global_file_path = create_resources_file_path(path_str)
        df = pd.read_csv(csv_global_file_path)
        self.gen_opt.design_variables = RoxieGeneticOptimization.initialize_design_variables(df)
        pop = [[1, 0, 1, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1,
                1, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1,
                0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1, 1,
                0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0]]

        chromosome = self.gen_opt.decode_individual_old(pop[0])
        geometry_rel = self.gen_opt.update_model_parameters_targeted(chromosome)

        # assert that the number of turns per layer is the same
        assert 13 == sum([block.block_def.nco for block in geometry_rel.blocks[:3]])
        assert 19 == sum([block.block_def.nco for block in geometry_rel.blocks[3:6]])
        assert 29 == sum([block.block_def.nco for block in geometry_rel.blocks[6:9]])
        assert 39 == sum([block.block_def.nco for block in geometry_rel.blocks[9:]])

    def test_split_list_into_chunks(self):
        # To produce a partition of t into k values:
        #
        # - Generate k-1 uniformly distributed values in the range [0, t].
        #
        # - Sort them, and add 0 at the beginning and t at the end.
        #
        # - Use the adjacent differences as the partition.
        np.random.seed(0)
        t = 30
        k = 4
        partition = RoxieGeneticOptimization.generate_number_random_partition(t, k)

        assert [13, 8, 6, 3] == partition

    def test_generate_chromosome_varying_turns_blocks(self):
        path_str = 'resources/optimization/design_variables_programmable_geometry.csv'
        csv_global_file_path = create_resources_file_path(path_str)
        df = pd.read_csv(csv_global_file_path)
        design_variables = RoxieGeneticOptimization.initialize_design_variables(df)

        chromosome = RoxieGeneticOptimization.generate_random_chromosome_programmable_geometry(design_variables)

        assert len(design_variables) - 8 == len(chromosome) // 6

    def test_decode_chromosome_varying_turns_blocks(self):
        path_str = 'resources/optimization/design_variables_programmable_geometry.csv'
        csv_global_file_path = create_resources_file_path(path_str)
        df = pd.read_csv(csv_global_file_path)
        design_variables = RoxieGeneticOptimization.initialize_design_variables(df)

        chromosome = RoxieGeneticOptimization.generate_random_chromosome_programmable_geometry(design_variables)

        self.gen_opt.design_variables = design_variables
        decoded = self.gen_opt.decode_chromosome_programmable_geometry(chromosome)

        assert decoded['spar_thickness:1:'] == design_variables[0].convert_gene_to_value(chromosome[0:6])
        assert decoded['nco:1:1'] == GeneticDesignVariable.convert_gene_to_int(chromosome[6:12])
        assert decoded['nco:1:2'] == GeneticDesignVariable.convert_gene_to_int(chromosome[12:18])
        assert decoded['nco:1:3'] == GeneticDesignVariable.convert_gene_to_int(chromosome[18:24])
        assert decoded['nco:1:4'] == GeneticDesignVariable.convert_gene_to_int(chromosome[24:30])
        assert decoded['nco:1:5'] == GeneticDesignVariable.convert_gene_to_int(chromosome[30:36])
        assert decoded['nco:1:6'] == GeneticDesignVariable.convert_gene_to_int(chromosome[36:42])
        assert decoded['phi_r:1:2'] == design_variables[9].convert_gene_to_value(chromosome[42:48])
        assert decoded['phi_r:1:3'] == design_variables[10].convert_gene_to_value(chromosome[48:54])
        assert decoded['phi_r:1:4'] == design_variables[11].convert_gene_to_value(chromosome[54:60])
        assert decoded['phi_r:1:5'] == design_variables[12].convert_gene_to_value(chromosome[60:66])
        assert decoded['phi_r:1:6'] == design_variables[13].convert_gene_to_value(chromosome[66:72])
        assert decoded['alpha_rad_r:1:2'] == design_variables[14].convert_gene_to_value(chromosome[72:78])
        assert decoded['alpha_rad_r:1:3'] == design_variables[15].convert_gene_to_value(chromosome[78:84])
        assert decoded['alpha_rad_r:1:4'] == design_variables[16].convert_gene_to_value(chromosome[84:90])
        assert decoded['alpha_rad_r:1:5'] == design_variables[17].convert_gene_to_value(chromosome[90:96])
        assert decoded['alpha_rad_r:1:6'] == design_variables[18].convert_gene_to_value(chromosome[96:102])
        assert decoded['spar_thickness:2:'] == design_variables[19].convert_gene_to_value(chromosome[102:108])
        assert decoded['nco:2:1'] == GeneticDesignVariable.convert_gene_to_int(chromosome[108:114])
        assert decoded['nco:2:2'] == GeneticDesignVariable.convert_gene_to_int(chromosome[114:120])
        assert decoded['nco:2:3'] == GeneticDesignVariable.convert_gene_to_int(chromosome[120:126])
        assert decoded['nco:2:4'] == GeneticDesignVariable.convert_gene_to_int(chromosome[126:132])
        assert decoded['nco:2:5'] == GeneticDesignVariable.convert_gene_to_int(chromosome[132:138])
        assert decoded['phi_r:2:2'] == design_variables[27].convert_gene_to_value(chromosome[138:144])
        assert decoded['phi_r:2:3'] == design_variables[28].convert_gene_to_value(chromosome[144:150])
        assert decoded['phi_r:2:4'] == design_variables[29].convert_gene_to_value(chromosome[150:156])
        assert decoded['phi_r:2:5'] == design_variables[30].convert_gene_to_value(chromosome[156:162])
        assert decoded['alpha_rad_r:2:2'] == design_variables[31].convert_gene_to_value(chromosome[162:168])
        assert decoded['alpha_rad_r:2:3'] == design_variables[32].convert_gene_to_value(chromosome[168:174])
        assert decoded['alpha_rad_r:2:4'] == design_variables[33].convert_gene_to_value(chromosome[174:180])
        assert decoded['alpha_rad_r:2:5'] == design_variables[34].convert_gene_to_value(chromosome[180:186])
        assert decoded['spar_thickness:3:'] == design_variables[35].convert_gene_to_value(chromosome[186:192])
        assert decoded['nco:3:1'] == GeneticDesignVariable.convert_gene_to_int(chromosome[192:198])
        assert decoded['nco:3:2'] == GeneticDesignVariable.convert_gene_to_int(chromosome[198:204])
        assert decoded['nco:3:3'] == GeneticDesignVariable.convert_gene_to_int(chromosome[204:210])
        assert decoded['nco:3:4'] == GeneticDesignVariable.convert_gene_to_int(chromosome[210:216])
        assert decoded['nco:3:5'] == GeneticDesignVariable.convert_gene_to_int(chromosome[216:222])
        assert decoded['phi_r:3:2'] == design_variables[43].convert_gene_to_value(chromosome[222:228])
        assert decoded['phi_r:3:3'] == design_variables[44].convert_gene_to_value(chromosome[228:234])
        assert decoded['phi_r:3:4'] == design_variables[45].convert_gene_to_value(chromosome[234:240])
        assert decoded['phi_r:3:5'] == design_variables[46].convert_gene_to_value(chromosome[240:246])
        assert decoded['alpha_rad_r:3:2'] == design_variables[47].convert_gene_to_value(chromosome[246:252])
        assert decoded['alpha_rad_r:3:3'] == design_variables[48].convert_gene_to_value(chromosome[252:258])
        assert decoded['alpha_rad_r:3:4'] == design_variables[49].convert_gene_to_value(chromosome[258:264])
        assert decoded['alpha_rad_r:3:5'] == design_variables[50].convert_gene_to_value(chromosome[264:270])
        assert decoded['spar_thickness:4:'] == design_variables[51].convert_gene_to_value(chromosome[270:276])
        assert decoded['nco:4:1'] == GeneticDesignVariable.convert_gene_to_int(chromosome[276:282])
        assert decoded['nco:4:2'] == GeneticDesignVariable.convert_gene_to_int(chromosome[282:288])
        assert decoded['nco:4:3'] == GeneticDesignVariable.convert_gene_to_int(chromosome[288:294])
        assert decoded['phi_r:4:2'] == design_variables[57].convert_gene_to_value(chromosome[294:300])
        assert decoded['phi_r:4:3'] == design_variables[58].convert_gene_to_value(chromosome[300:306])
        assert decoded['alpha_rad_r:4:2'] == design_variables[59].convert_gene_to_value(chromosome[306:312])
        assert decoded['alpha_rad_r:4:3'] == design_variables[60].convert_gene_to_value(chromosome[312:318])

    def test_update_model_varying_turns_blocks(self):
        np.random.seed(1)
        # initialize a slotted geometry_rel model
        json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_rel_slotted.json')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)
        geometry_rel = GeometryFactory.init_slotted_with_json(json_path, cadata)

        # initialize design variables
        path_str = 'resources/optimization/design_variables_programmable_geometry.csv'
        csv_global_file_path = create_resources_file_path(path_str)
        df = pd.read_csv(csv_global_file_path)
        design_variables = RoxieGeneticOptimization.initialize_design_variables(df)

        chromosome = RoxieGeneticOptimization.generate_random_chromosome_programmable_geometry(design_variables)

        self.gen_opt.design_variables = design_variables
        decoded = self.gen_opt.decode_chromosome_programmable_geometry(chromosome)

        self.gen_opt.geometry = geometry_rel
        geometry_rel = self.gen_opt.update_model_parameters_programmable_geometry(decoded)

        # assert updated spar_thickness
        assert geometry_rel.layer_defs[0].spar_thickness == decoded['spar_thickness:1:']
        assert geometry_rel.layer_defs[1].spar_thickness == decoded['spar_thickness:2:']
        assert geometry_rel.layer_defs[2].spar_thickness == decoded['spar_thickness:3:']
        assert geometry_rel.layer_defs[3].spar_thickness == decoded['spar_thickness:4:']

        # assert nco blocks
        assert geometry_rel.blocks[0].block_def.nco == decoded['nco:1:1']
        assert geometry_rel.blocks[1].block_def.nco == decoded['nco:1:2']
        assert geometry_rel.blocks[2].block_def.nco == decoded['nco:1:3']
        assert abs(geometry_rel.blocks[0].block_def.phi_r - 0.5729399999999999) < 1e-6
        assert abs(geometry_rel.blocks[1].block_def.phi_r - 2.076404240564713) < 1e-6
        assert abs(geometry_rel.blocks[2].block_def.phi_r - 4.793637355826874) < 1e-6
        assert geometry_rel.layer_defs[0].blocks == [1, 2, 3]

        assert geometry_rel.blocks[3].block_def.nco == decoded['nco:2:1']
        assert geometry_rel.blocks[4].block_def.nco == decoded['nco:2:2']
        assert geometry_rel.blocks[5].block_def.nco == decoded['nco:2:3']
        assert geometry_rel.blocks[6].block_def.nco == decoded['nco:2:4']
        assert abs(geometry_rel.blocks[3].block_def.phi_r - 0.36728000000000005) < 1e-6
        assert abs(geometry_rel.blocks[4].block_def.phi_r - 2.6875) < 1e-6
        assert abs(geometry_rel.blocks[5].block_def.phi_r - 4.6522330804394585) < 1e-6
        assert abs(geometry_rel.blocks[6].block_def.phi_r - 2.9632517062888084) < 1e-6
        assert geometry_rel.layer_defs[1].blocks == [4, 5, 6, 7]

        assert geometry_rel.blocks[7].block_def.nco == decoded['nco:3:1']
        assert geometry_rel.blocks[8].block_def.nco == decoded['nco:3:2']
        assert geometry_rel.blocks[9].block_def.nco == decoded['nco:3:3']
        assert geometry_rel.blocks[10].block_def.nco == decoded['nco:3:4']
        assert geometry_rel.blocks[11].block_def.nco == decoded['nco:3:5']
        assert abs(geometry_rel.blocks[7].block_def.phi_r - 0.2702600000000001) < 1e-6
        assert abs(geometry_rel.blocks[8].block_def.phi_r - 1.53125) < 1e-6
        assert abs(geometry_rel.blocks[9].block_def.phi_r - 3.2932184171452334) < 1e-6
        assert abs(geometry_rel.blocks[10].block_def.phi_r - 1.8359336038394076) < 1e-6
        assert abs(geometry_rel.blocks[11].block_def.phi_r - 3.222769077299759) < 1e-6
        assert geometry_rel.layer_defs[2].blocks == [8, 9, 10, 11, 12]

        assert geometry_rel.blocks[12].block_def.nco == decoded['nco:4:1']
        assert geometry_rel.blocks[13].block_def.nco == decoded['nco:4:2']
        assert geometry_rel.blocks[14].block_def.nco == decoded['nco:4:3']
        assert abs(geometry_rel.blocks[12].block_def.phi_r - 0.21236) < 1e-6
        assert abs(geometry_rel.blocks[13].block_def.phi_r - 2.84375) < 1e-6
        assert abs(geometry_rel.blocks[14].block_def.phi_r - 1.9686957925839579) < 1e-6
        assert geometry_rel.layer_defs[3].blocks == [13, 14, 15]



from magnumapi.geometry.blocks.CosThetaBlock import CosThetaBlock
from magnumapi.geometry.primitives.Area import Area
from magnumapi.geometry.primitives.Line import Line


class GeometryCheck:
    @staticmethod
    def is_outside_of_first_quadrant(geometry, eps=1e-30) -> True:
        """ Method checking whether a geometry is fully positioned in the first quadrant.

        :param eps: a machine precision for comparison of 0 value
        :return: True if an Area instance is in the first quadrant, False otherwise
        """
        for layer_def in geometry.layer_defs:
            for block_index in layer_def.blocks:
                if geometry.blocks[block_index - 1].is_outside_of_first_quadrant(eps=eps):
                    return True

        return False

    @staticmethod
    def are_turns_overlapping(geometry) -> bool:
        """Method checking whether turns from neighbouring blocks are overlapping.

        :return: True, if at least two neighbouring turns are overlapping, otherwise False
        """
        for layer_def in geometry.layer_defs:
            for block_index in layer_def.blocks[1:]:
                area_prev = geometry.blocks[block_index - 2].areas[-1]
                area_curr = geometry.blocks[block_index - 1].areas[0]

                if Area.are_areas_overlapping(area_prev, area_curr):
                    return True

        return False

    @classmethod
    def is_wedge_tip_too_sharp(cls, geometry, min_value_in_mm: float) -> bool:
        """ Method checking whether wedge tip is too sharp. Method computes the wedge tip arc length, by finding an
        intersection of two neighbouring turns with an inner radius

        :param min_value_in_mm: minimum value of  the wedge tip arc
        :return: True if wedge length is below the minimum value
        """
        for layer_def in geometry.layer_defs:
            for block_index in layer_def.blocks[1:]:
                block_curr = geometry.blocks[block_index - 1]
                block_prev = geometry.blocks[block_index - 2]
                if isinstance(block_curr, CosThetaBlock) and isinstance(block_prev, CosThetaBlock):
                    if cls.is_wedge_arc_of_two_blocks_too_short(block_curr, block_prev, min_value_in_mm):
                        return True

        return False

    @staticmethod
    def is_wedge_arc_of_two_blocks_too_short(block_curr: "CosThetaBlock",
                                             block_prev: "CosThetaBlock",
                                             min_value_in_mm: float) -> bool:
        """ Static method calculating a wedge arc length between two neighbouring blocks of a given layer

        :param block_curr: current block
        :param block_prev: previous block
        :param min_value_in_mm: minimum arc length in millimeters
        :return: True if the wedge arc length is too short, otherwise False
        """
        area_curr = block_curr.areas[0]
        area_prev = block_prev.areas[-1]

        line_first_area = area_curr.get_line(0)
        line_last_area = area_prev.get_line(2)

        radius = block_curr.block_def.radius
        l_arc = Line.calc_arc_length_between_two_lines(radius, line_last_area, line_first_area)

        return l_arc < min_value_in_mm



from dataclasses import dataclass

from magnumapi.cadata.Definition import Definition


@dataclass
class TransientDefinition(Definition):
    """Class for transient definition.

       Attributes:
           r_c (float): The cross resistance (Ohm) between strands.
           r_a (float): The adjacent resistance (Ohm) between strands.
           l_fil_tp (float): The filament twist pitch length (m).
           res_0 (float: The constant part (Ohm/m) of the magnetoresistive matrix copper
           dres_over_db (float): The derivative dR/dB (Ohm/mT) of the magnetoresistive matrix copper.
           f_strand_fill (float): The filling factor of a strand for IFCC calculations
    """
    r_c: float = 0.0
    r_a: float = 0.0
    l_fil_tp: float = 0.0
    res_0: float = 0.0
    dres_over_db: float = 0.0
    f_strand_fill: float = 0.0

    @staticmethod
    def get_magnum_to_roxie_dct() -> dict:
        return {"name": "Name",
                "r_c": "Rc",
                "r_a": "Ra",
                "l_fil_tp": "fil.twistp.",
                "res_0": "fil.R0",
                "dres_over_db": "fil.dR/dB",
                "f_strand_fill": "strandfill.fac.",
                "comment": "Comment"}



from typing import List

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import matplotlib.pyplot as plt
from ipyaggrid import Grid

from magnumapi.geometry.blocks.Block import Block
from magnumapi.geometry.primitives.Arc import Arc


class GeometryPlot:
    @classmethod
    def plotly_geometry_blocks(cls, geometry, figsize=(750, 750), xlim=(0, 80), ylim=(0, 80)):
        """ Method plotting blocks with plotly library

        :param figsize: size of the figure on a display
        :param xlim: limits in x-direction
        :param ylim: limits in y-direction
        """
        go_scatters = cls._create_plotly_scatters(geometry.blocks)

        fig = go.Figure(go_scatters)
        fig.update_layout(
            autosize=False,
            width=figsize[0],
            height=figsize[1],
            yaxis_range=ylim,
            xaxis_range=xlim,
            showlegend=False,
            plot_bgcolor='rgba(0,0,0,0)'
        )
        fig.add_trace(px.scatter(x=[0], y=[0]).data[0])
        fig.update_xaxes(title=dict(text='x, [mm]'))
        fig.update_yaxes(title=dict(text='y, [mm]'))
        fig.show()

    @staticmethod
    def _create_plotly_scatters(blocks: List[Block]) -> List[go.Scatter]:
        """ Static method creating a list of plotly scatter instances

        :param blocks: list of blocks to convert to plotly scatter
        :return: a list of plotly scatter instances
        """
        go_scatter = []
        index = 1
        for block in blocks:
            for area in block.areas:
                x = [line.p1.x for line in area.lines]
                y = [line.p1.y for line in area.lines]
                x.append(area.get_line(0).p1.x)
                y.append(area.get_line(0).p1.y)

                go_scatter.append(go.Scatter(x=x, y=y, fill="toself", fillcolor='white', marker=dict(size=1),
                                             name='turn' + str(index), line=go.scatter.Line(color='blue')))
                index += 1

        return go_scatter

    @staticmethod
    def plot_blocks(geometry, figsize=(10, 10), is_grid=True, xlim=(0, 80), ylim=(0, 80)) -> None:
        """ Method plotting all blocks with matplotlib library

        :param figsize: size of a figure on a screen
        :param is_grid: if True, then the grid is displayed, otherwise False
        :param xlim: limits in x-direction
        :param ylim: limits in y-direction
        """
        fig, ax = plt.subplots(figsize=figsize)
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)
        ax.grid(is_grid)

        for block in geometry.blocks:
            block.plot_block(ax)

        plt.show()

    @staticmethod
    def plot_bare_blocks(geometry, figsize=(15, 15), is_grid=True, xlim=(0, 80), ylim=(0, 80)) -> None:
        """ Method plotting bare (uninsulated) areas for all blocks in a geometry

        :param figsize: size of a figure on a screen
        :param is_grid: if True, then the grid is displayed, otherwise False
        :param xlim: limits in x-direction
        :param ylim: limits in y-direction
        """
        fig, ax = plt.subplots(figsize=figsize)
        ax.set_aspect('equal', 'box')
        ax.grid(is_grid)
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)
        for block in geometry.blocks:
            block.plot_bare_block(ax)

        plt.show()

    @staticmethod
    def plotly_homogenized_geometry_blocks(geometry, figsize=(65, 65), xlim=(0, 80), ylim=(0, 80)):

        lines = []
        arcs = []

        index = 1
        for block in geometry.blocks:
            line_bottom = block.areas[0].get_line(0)
            arc_outer = block.areas[0].get_line(1)
            line_top = block.areas[0].get_line(2)
            arc_inner = block.areas[0].get_line(3)

            arcs.append(dict(type="path",
                             path=Arc.describe_ellipse_arc(a=arc_inner.get_radius(),
                                                           b=arc_inner.get_radius(),
                                                           start_angle=arc_inner.get_start_angle_in_rad(),
                                                           end_angle=arc_inner.get_end_angle_in_rad()),
                             line_color="black"))

            arcs.append(dict(type="path",
                             path=Arc.describe_ellipse_arc(a=arc_outer.get_radius(),
                                                           b=arc_outer.get_radius(),
                                                           start_angle=arc_outer.get_start_angle_in_rad(),
                                                           end_angle=arc_outer.get_end_angle_in_rad()),
                             line_color="black"))

            lines.append(go.Scatter(x=[line_bottom.p1.x, line_bottom.p2.x],
                                    y=[line_bottom.p1.y, line_bottom.p2.y],
                                    fill="toself", fillcolor='white', marker=dict(size=1),
                                    name='block' + str(index), line=go.scatter.Line(color='black')))

            lines.append(go.Scatter(x=[line_top.p1.x, line_top.p2.x],
                                    y=[line_top.p1.y, line_top.p2.y],
                                    fill="toself", fillcolor='white', marker=dict(size=1),
                                    name='block' + str(index), line=go.scatter.Line(color='black')))

            index += 1

        fig = go.Figure(lines)
        fig.update_layout(
            autosize=False,
            width=figsize[0],
            height=figsize[1],
            xaxis_range=xlim,
            yaxis_range=ylim,
            showlegend=False,
            plot_bgcolor='rgba(0,0,0,0)',
            shapes=arcs
        )
        fig.add_trace(px.scatter(x=[0], y=[0]).data[0])
        fig.update_xaxes(title=dict(text='x, [mm]'))
        fig.update_yaxes(title=dict(text='y, [mm]'))
        fig.show()

    @staticmethod
    def display_definition_table(df: pd.DataFrame) -> Grid:
        """ Static method displaying a definition table in an editable format in a notebook

        :param df: a dataframe to be displayed
        :return: an object to be displayed in a notebook
        """
        column_defs = [{'headername': c, 'field': c} for c in df.columns]

        grid_options = {
            'columnDefs': column_defs,
            'enableSorting': True,
            'enableFilter': True,
            'enableColResize': True,
            'enableRangeSelection': True,
            'rowSelection': 'multiple',
        }

        return Grid(grid_data=df,
                    grid_options=grid_options,
                    quick_filter=True,
                    show_toggle_edit=True,
                    sync_on_edit=True,
                    export_csv=True,
                    export_excel=True,
                    theme='ag-theme-balham',
                    show_toggle_delete=True,
                    columns_fit='auto',
                    index=False)



from dataclasses import dataclass

from magnumapi.geometry.definitions.RectangularBlockDefinition import RectangularBlockDefinition


@dataclass
class HomogenizedRectangularBlockDefinition(RectangularBlockDefinition):
    """Class for a homogenized rectangular block definition.

    Attributes:
       x_up_r (float): The x-coordinate of upper right block corner (assuming alpha equal to 0).
       y_up_r (float): The y-coordinate of upper right block corner (assuming alpha equal to 0).
    """
    x_up_r: float
    y_up_r: float



from dataclasses import dataclass

from magnumapi.cadata.Definition import Definition


@dataclass
class InsulationDefinition(Definition):
    """Class for insulation definition.

       Attributes:
           thickness (float): The thickness (mm) on narrow side of cable.
           width (float): The width (mm) on broad side of cable.
    """
    thickness: float
    width: float

    @staticmethod
    def get_magnum_to_roxie_dct() -> dict:
        return {"name": "Name",
                "thickness": "Radial",
                "width": "Azimut",
                "comment": "Comment"}



from dataclasses import dataclass
from typing import List
import magnumapi.commons.json_file as json_file

from magnumapi.optimization.OptimizationConfig import OptimizationConfig
from magnumapi.optimization.ObjectiveConfig import ObjectiveConfig
from magnumapi.optimization.OptimizationNotebookConfig import OptimizationNotebookConfig


@dataclass
class ParallelOptimizationConfig(OptimizationConfig):
    """Class for parallel optimization config used for the genetic algorithm.

    Attributes:
       ansys (ParallelAnsysConfig): parallel ansys config

    """
    ansys: "ParallelAnsysConfig"

    def __str__(self) -> str:
        notebooks_str = "\n\n".join(str(notebook) for notebook in self.notebooks)
        objectives_str = "\n\n".join(str(objective) for objective in self.objectives)
        return "input_folder: %s\n" \
               "output_folder: %s\n" \
               "logger_rel_path: %s\n" \
               "n_pop: %d\n" \
               "n_gen: %d\n" \
               "r_cross: %f\n" \
               "r_mut: %f\n" \
               "objectives: \n\n" \
               "%snotebooks: \n\n" \
               "%s \n\n" \
               "ansys: %s" % (self.input_folder,
                              self.output_folder,
                              self.logger_rel_path,
                              self.n_pop,
                              self.n_gen,
                              self.r_cross,
                              self.r_mut,
                              objectives_str,
                              notebooks_str,
                              self.ansys)

    @staticmethod
    def initialize_config(json_path: str) -> "OptimizationConfig":
        """ Static method initializing an optimization config from a json file

        :param json_path: a path to a json file with config
        :return: initialized OptimizationConfig instance
        """
        data = json_file.read(json_path)

        input_folder = data['input_folder']
        output_folder = data['output_folder']
        logger_rel_path = data['logger_rel_path']
        n_pop = data['n_pop']
        n_gen = data['n_gen']
        r_cross = data['r_cross']
        r_mut = data['r_mut']
        objectives = [ObjectiveConfig(**ff) for ff in data['objectives']]
        notebooks = [OptimizationNotebookConfig(**nb) for nb in data['notebooks']]
        ansys = ParallelAnsysConfig(**data['ansys'])

        return ParallelOptimizationConfig(input_folder=input_folder,
                                          output_folder=output_folder,
                                          logger_rel_path=logger_rel_path,
                                          n_pop=n_pop,
                                          n_gen=n_gen,
                                          r_cross=r_cross,
                                          r_mut=r_mut,
                                          objectives=objectives,
                                          notebooks=notebooks,
                                          ansys=ansys)

    def get_weight(self, objective: str) -> float:
        """ Method iterating through the list of objective config, finding a matching name and returning
        corresponding weight.

        :param objective: name of an objective variable
        :return: value of the objective weight
        """
        for config_objective in self.objectives:
            if config_objective.objective == objective:
                return config_objective.weight

        raise KeyError('Objective name %s not present in objective configs.' % objective)

    def get_constraint(self, objective: str) -> float:
        """ Method iterating through the list of objective config, finding a matching name and returning
        corresponding constraint.

        :param objective: name of an objective variable
        :return: value of the objective constraint
        """
        for config_objective in self.objectives:
            if config_objective.objective == objective:
                return config_objective.constraint

        raise KeyError('Objective name %s not present in objective configs.' % objective)


@dataclass
class ParallelAnsysConfig:
    run_dir: str
    n_parallel_runners: int
    n_proc: int
    root_dir: str



from unittest import TestCase

import numpy as np

from magnumapi.optimization.DesignVariable import DesignVariable, GeneticDesignVariable


class TestDesignVariable(TestCase):
    def test_get_variable_name_empty_bcs(self):
        # arrange
        dv = DesignVariable(xl=5, xu=10, variable='R_EE', layer='', bcs='')

        # act
        variable_names = dv.get_variable_names()

        # assert
        self.assertEqual('R_EE', variable_names[0])

    def test_get_variable_name_single_index_bcs(self):
        # arrange
        dv = DesignVariable(xl=5, xu=10, variable='nco', layer=1, bcs='1')

        # act
        variable_names = dv.get_variable_names()

        # assert
        self.assertEqual('nco:1:1', variable_names[0])

    def test_get_variable_name_range_index_bcs(self):
        # arrange
        dv = DesignVariable(xl=5, xu=10, variable='nco', layer=1, bcs='1-5')

        # act
        variable_names = dv.get_variable_names()

        # assert
        self.assertListEqual(['nco:1:1', 'nco:1:2', 'nco:1:3', 'nco:1:4', 'nco:1:5'], variable_names)

    def test_get_variable_name_error(self):
        # arrange
        dv = DesignVariable(xl=5, xu=10, variable='nco', layer=1, bcs='1:5')

        # act
        # assert
        with self.assertRaises(AttributeError) as context:
            dv.get_variable_names()

        self.assertTrue('The design variable has incorrect block index value: 1:5.' in str(context.exception))

    def test__convert_range_of_blocks_into_list_error_two_hyphens(self):
        # arrange
        dv = DesignVariable(xl=5, xu=10, variable='nco', layer=1, bcs='1--')

        # act
        # assert
        with self.assertRaises(AttributeError) as context:
            dv._convert_range_of_blocks_into_list()

        self.assertTrue('The block index range definition 1-- is wrong. Only one hyphen is allowed.'
                        in str(context.exception))

    def test__convert_range_of_blocks_into_list_error_lower_limit_character(self):
        # arrange
        dv = DesignVariable(xl=5, xu=10, variable='nco', layer=1, bcs='a-5')

        # act
        # assert
        with self.assertRaises(AttributeError) as context:
            dv._convert_range_of_blocks_into_list()

        self.assertTrue('The lower block index range a is not a number.' in str(context.exception))

    def test__convert_range_of_blocks_into_list_error_upper_limit_character(self):
        # arrange
        dv = DesignVariable(xl=5, xu=10, variable='nco', layer=1, bcs='5-b')

        # act
        # assert
        with self.assertRaises(AttributeError) as context:
            dv._convert_range_of_blocks_into_list()

        self.assertTrue('The upper block index range b is not a number.' in str(context.exception))

    def test__convert_range_of_blocks_into_list_lower_greater_than_upper(self):
        # arrange
        dv = DesignVariable(xl=5, xu=10, variable='nco', layer=1, bcs='5-1')

        # act
        # assert
        with self.assertRaises(AttributeError) as context:
            dv._convert_range_of_blocks_into_list()

        self.assertTrue('The lower index 5 is greater than the upper one 1.' in str(context.exception))

    def test__convert_range_of_blocks_into_list_lower_equal_to_upper(self):
        # arrange
        dv = DesignVariable(xl=5, xu=10, variable='nco', layer=1, bcs='5-5')

        # act
        variable_name = dv._convert_range_of_blocks_into_list()

        # assert
        self.assertListEqual(['nco:1:5'], variable_name)

    def test__convert_range_of_blocks_into_list(self):
        # arrange
        dv = DesignVariable(xl=5, xu=10, variable='nco', layer=1, bcs='1-5')

        # act
        variable_name = dv._convert_range_of_blocks_into_list()

        # assert
        self.assertListEqual(['nco:1:1', 'nco:1:2', 'nco:1:3', 'nco:1:4', 'nco:1:5'], variable_name)


class TestGeneticDesignVariable(TestCase):
    def test_generate_random_gene_int_seed_0(self):
        # arrange
        np.random.seed(0)
        dv = GeneticDesignVariable(xl=5, xu=10, variable='nco', layer=1, bcs='1-5', bits=5, variable_type='int')

        # act
        gene = dv.generate_random_gene()

        # assert
        self.assertListEqual([0, 0, 1, 0, 1], gene)

    def test_generate_random_gene_int_seed_2001(self):
        # arrange
        np.random.seed(2001)
        dv = GeneticDesignVariable(xl=5, xu=10, variable='nco', layer=1, bcs='1-5', bits=4, variable_type='int')

        # act
        gene = dv.generate_random_gene()

        # assert
        self.assertListEqual([0, 0, 0, 0], gene)

    def test_generate_random_gene_float_seed_0(self):
        # arrange
        np.random.seed(0)
        dv = GeneticDesignVariable(xl=5, xu=10, variable='nco', layer=1, bcs='1-5', bits=5, variable_type='float')

        # act
        gene = dv.generate_random_gene()

        # assert
        self.assertListEqual([0, 1, 1, 0, 1], gene)

    def test_convert_gene_to_value_int(self):
        # arrange
        np.random.seed(0)
        dv = GeneticDesignVariable(xl=5, xu=10, variable='nco', layer=1, bcs='1-5', bits=5, variable_type='int')

        # act
        gene = dv.generate_random_gene()
        value = dv.convert_gene_to_value(gene)

        # assert
        self.assertEqual(10, value)

    def test_convert_gene_to_value_float(self):
        # arrange
        np.random.seed(0)
        dv = GeneticDesignVariable(xl=5, xu=10, variable='nco', layer=1, bcs='1-5', bits=5, variable_type='float')

        # act
        gene = dv.generate_random_gene()
        value = dv.convert_gene_to_value(gene)

        # assert
        self.assertAlmostEqual(7.03125, value, places=5)

    def test_convert_gene_to_int(self):
        # arrange
        gene = [1, 0, 1, 0]

        # act
        int_value = GeneticDesignVariable.convert_gene_to_int(gene)

        # assert
        self.assertEqual(10, int_value)

    def test_convert_int_to_gene(self):
        # arrange
        int_value = 10

        # act
        gene = GeneticDesignVariable.convert_int_to_gene(int_value, 4)

        # assert
        self.assertListEqual([1, 0, 1, 0], gene)

    def test_convert_int_to_gene_five_bits(self):
        # arrange
        int_value = 10

        # act
        gene = GeneticDesignVariable.convert_int_to_gene(int_value, 5)

        # assert
        self.assertListEqual([0, 1, 0, 1, 0], gene)

class TestCockpitGeneticDesignVariable(TestCase):

    def test(self):
        pass




import unittest

import numpy as np

from magnumapi.matpro.critical_currents import calc_jc_nbti_bottura, calc_jc_nb3sn_bordini, calc_jc_nb3sn_summers, \
    calc_jc_nb3sn_summers_orig
from tests.resource_files import create_resources_file_path


class Test_critical_currents(unittest.TestCase):
    def test_calc_jc_nbti_bottura(self):
        # arrange
        T_0 = 1.9
        T = np.linspace(T_0, 8, 100)

        # act
        jc_nbti_bottura = calc_jc_nbti_bottura(T, B=1.0)

        # assert
        path = create_resources_file_path('resources/matpro/critical_currents/jc_nbti_bottura.txt')
        jc_nbti_bottura_ref = np.loadtxt(path)
        np.testing.assert_allclose(jc_nbti_bottura_ref, jc_nbti_bottura)

    def test_calc_jc_nb3sn_bordini(self):
        # arrange
        T_0 = 1.9
        T = np.linspace(T_0, 8, 100)

        # act
        jc_nb3sn_bordini = calc_jc_nb3sn_bordini(T, B=1.0)

        # assert
        path = create_resources_file_path('resources/matpro/critical_currents/jc_nb3sn_bordini.txt')
        jc_nb3sn_bordini_ref = np.loadtxt(path)
        np.testing.assert_allclose(jc_nb3sn_bordini_ref, jc_nb3sn_bordini)

    def test_calc_jc_nb3sn_summers(self):
        # arrange
        T_0 = 1.9
        T = np.linspace(T_0, 8, 100)

        # act
        jc_nb3sn_summers = calc_jc_nb3sn_summers(T, B=1.0)

        # assert
        path = create_resources_file_path('resources/matpro/critical_currents/jc_nb3sn_summers.txt')
        jc_nb3sn_summers_ref = np.loadtxt(path)
        np.testing.assert_allclose(jc_nb3sn_summers_ref, jc_nb3sn_summers)

    def test_calc_jc_nb3sn_summers_orig(self):
        # arrange
        T_0 = 1.9
        T = np.linspace(T_0, 8, 100)

        # act
        jc_nb3sn_summers_orig = calc_jc_nb3sn_summers_orig(T, B=1.0)

        # assert
        path = create_resources_file_path('resources/matpro/critical_currents/jc_nb3sn_summers_orig.txt')
        jc_nb3sn_summers_orig_ref = np.loadtxt(path)
        np.testing.assert_allclose(jc_nb3sn_summers_orig_ref, jc_nb3sn_summers_orig)

if __name__ == '__main__':
    unittest.main()



import scooby


class Report(scooby.Report):
    """Class for custom scooby.Report."""

    def __init__(self, additional=None, ncol=3, text_width=80, sort=False) -> None:
        # Mandatory packages
        core = ['numpy', 'pandas', 'numpy', 'dataclasses', 'ipyaggrid', 'plotly', 'ipyaggrid', 'scrapbook', 'papermill',
                'scooby', 'ansys.mapdl.core', 'ansys.mapdl.reader']
        optional = ['matplotlib', 'coverage']

        scooby.Report.__init__(self,
                               additional=additional,
                               core=core,
                               optional=optional,
                               ncol=ncol,
                               text_width=text_width, sort=sort)



import os

from magnumapi.tool_adapters.roxie.RoxieToolAdapter import TerminalRoxieToolAdapter

config = {
    "executable_name": "runroxie",
    "input_folder_rel_dir": "input",
    "input_file": "input.data",
    "output_file": "input.output",
    "cadata_file": "roxieold_2.cadata",
    "xml_output_file": "roxieData.xml"
}

def test_constructor():
    # arrange

    # act
    roxie = TerminalRoxieToolAdapter(**config)

    # assert
    assert 'runroxie' == roxie.executable_name
    assert 'input' == roxie.input_folder_rel_dir
    assert 'input%sroxieData.xml' % os.sep == roxie.xml_output_file_path
    assert 'input%sinput.output' % os.sep == roxie.output_file_path
    assert 'input%sinput.data' % os.sep == roxie.input_file_path
    assert 'input%sroxieold_2.cadata' % os.sep == roxie.cadata_file_path
    assert 'roxieold_2.cadata' == roxie.cadata_file

def test__get_process_command():
    # arrange

    # act
    roxie = TerminalRoxieToolAdapter(**config)
    command = roxie._get_process_command()

    assert  ['runroxie', 'input%sinput.data' % os.sep] == command



from dataclasses import dataclass

from magnumapi.cadata.Definition import Definition


@dataclass
class RemFitDefinition(Definition):
    """Class for remenent magnetization definition definition.

       Attributes:
           type (float): The Remfit type.
           c1 (float): fit coefficient 1.
           c2 (float): fit coefficient 2.
           c3 (float): fit coefficient 3.
           c4 (float): fit coefficient 4.
           c5 (float): fit coefficient 5.
           c6 (float): fit coefficient 6.
           c7 (float): fit coefficient 7.
           c8 (float): fit coefficient 8.
           c9 (float): fit coefficient 9.
           c10 (float): fit coefficient 10.
           c11 (float): fit coefficient 11.
    """
    type: str = ''
    c1: float = 0.0
    c2: float = 0.0
    c3: float = 0.0
    c4: float = 0.0
    c5: float = 0.0
    c6: float = 0.0
    c7: float = 0.0
    c8: float = 0.0
    c9: float = 0.0
    c10: float = 0.0
    c11: float = 0.0

    @staticmethod
    def get_magnum_to_roxie_dct() -> dict:
        return {"name": "Name",
                "type": "Type",
                "c1": "C1",
                "c2": "C2",
                "c3": "C3",
                "c4": "C4",
                "c5": "C5",
                "c6": "C6",
                "c7": "C7",
                "c8": "C8",
                "c9": "C9",
                "c10": "C10",
                "c11": "C11",
                "comment": "Comment"}



from dataclasses import dataclass

from magnumapi.geometry.definitions.BlockDefinition import BlockDefinition


@dataclass
class CosThetaBlockDefinition(BlockDefinition):
    """Class for cos-theta block definition.

    Attributes:
       radius (float): The radius in millimeter.
    """
    radius: float



import os
import pytest

from magnumapi.tool_adapters.ToolAdapterFactory import ToolAdapterFactory
from magnumapi.tool_adapters.roxie.RoxieToolAdapter import TerminalRoxieToolAdapter
from magnumapi.tool_adapters.roxie.RoxieToolAdapter import DockerTerminalRoxieToolAdapter
from tests.resource_files import create_resources_file_path


def test_init_with_json_terminal_roxie():
    # arrange
    config_path = create_resources_file_path('resources/tool_adapters/roxie/terminal_roxie.json')

    # act
    tool_adapter = ToolAdapterFactory.init_with_json(config_path)

    # assert
    assert isinstance(tool_adapter, TerminalRoxieToolAdapter)
    assert 'input' == tool_adapter.input_folder_rel_dir
    assert 'runroxie' == tool_adapter.executable_name
    assert 'input%sroxieData.xml' % os.sep == tool_adapter.xml_output_file_path
    assert 'input%sinput.output' % os.sep == tool_adapter.output_file_path
    assert 'input%sinput.data' % os.sep == tool_adapter.input_file_path
    assert 'input%sroxieold_2.cadata' % os.sep == tool_adapter.cadata_file_path
    assert 'roxieold_2.cadata' == tool_adapter.cadata_file


def test_init_with_json_docker_terminal_roxie():
    # arrange
    config_path = create_resources_file_path('resources/tool_adapters/roxie/docker_terminal_roxie.json')

    # act
    tool_adapter = ToolAdapterFactory.init_with_json(config_path)

    # assert
    assert isinstance(tool_adapter, DockerTerminalRoxieToolAdapter)
    assert 'input' == tool_adapter.input_folder_rel_dir
    assert 'runroxie' == tool_adapter.executable_name
    assert 'input%sroxieData.xml' % os.sep == tool_adapter.xml_output_file_path
    assert 'input%sinput.output' % os.sep == tool_adapter.output_file_path
    assert 'input%sinput.data' % os.sep == tool_adapter.input_file_path
    assert 'input%sroxieold_2.cadata' % os.sep == tool_adapter.cadata_file_path
    assert 'runroxie' == tool_adapter.executable_name
    assert 'roxieold_2.cadata' == tool_adapter.cadata_file


def test_init_with_dict_terminal_roxie():
    # arrange
    config = {
        "executable_name": "runroxie",
        "input_folder_rel_dir": "input",
        "input_file": "input.data",
        "output_file": "input.output",
        "cadata_file": "roxieold_2.cadata",
        "xml_output_file": "roxieData.xml"
    }

    # act
    tool_adapter = ToolAdapterFactory.init_with_dict(config)

    # assert
    assert isinstance(tool_adapter, TerminalRoxieToolAdapter)
    assert 'input' == tool_adapter.input_folder_rel_dir
    assert 'runroxie' == tool_adapter.executable_name
    assert 'input%sroxieData.xml' % os.sep == tool_adapter.xml_output_file_path
    assert 'input%sinput.output' % os.sep == tool_adapter.output_file_path
    assert 'input%sinput.data' % os.sep == tool_adapter.input_file_path
    assert 'input%sroxieold_2.cadata' % os.sep == tool_adapter.cadata_file_path
    assert 'roxieold_2.cadata' == tool_adapter.cadata_file


def test_init_with_dict_docker_terminal_roxie():
    # arrange
    config = {
        "executable_name": "runroxie",
        "input_folder_rel_dir": "input",
        "input_file": "input.data",
        "output_file": "input.output",
        "cadata_file": "roxieold_2.cadata",
        "xml_output_file": "roxieData.xml",
        "docker_image_name": "roxie_terminal"
    }

    # act
    tool_adapter = ToolAdapterFactory.init_with_dict(config)

    # assert
    assert isinstance(tool_adapter, DockerTerminalRoxieToolAdapter)
    assert 'input' == tool_adapter.input_folder_rel_dir
    assert 'runroxie' == tool_adapter.executable_name
    assert 'input%sroxieData.xml' % os.sep == tool_adapter.xml_output_file_path
    assert 'input%sinput.output' % os.sep == tool_adapter.output_file_path
    assert 'input%sinput.data' % os.sep == tool_adapter.input_file_path
    assert 'input%sroxieold_2.cadata' % os.sep == tool_adapter.cadata_file_path
    assert 'roxieold_2.cadata' == tool_adapter.cadata_file


def test_init_with_dict_docker_terminal_roxie_error():
    # arrange
    config = {
        "executable_name": "runroxie",
        "input_folder_rel_dir": "input",
        "input_file": "input.data",
        "output_file": "input.output",
        "cadata_file": "roxieold_2.cadata",
        "xml_output_file": "roxieData.xml",
        "docker_image": "roxie_terminal"
    }

    # act
    with pytest.raises(KeyError) as exc_info:
        tool_adapter = ToolAdapterFactory.init_with_dict(config)

    assert 'The input config definition keys dict_keys([\'executable_name\', \'input_folder_rel_dir\', ' \
           '\'input_file\', \'output_file\', \'cadata_file\', \'xml_output_file\', \'docker_image\']) ' \
           'do not match any tool adapter constructor signature!' in str(exc_info.value)


def test_init_with_dict_terminal_ansys():
    # arrange
    config = {"input_folder_rel_dir": "input\\Mechanic_Plate",
              "input_file": "15T_mech.inp",
              "output_file": "vallone.out",
              "model_file": "Model.inp",
              "rst_file": "15T_2d_mech.rst",
              "out_temp_file": "out_tmp_mech.txt",
              "license_type": "-aa_r",
              "executable_path": "C:\\Program Files\\ANSYS Inc\\v202\\ansys\\bin\\winx64\\ANSYS202.exe"}

    # act
    tool_adapter = ToolAdapterFactory.init_with_dict(config)

    # assert
    assert os.getcwd() == tool_adapter.root_dir
    assert "input\\Mechanic_Plate" == tool_adapter.input_folder_rel_dir
    assert "15T_mech.inp" == tool_adapter.input_file
    assert "vallone.out" == tool_adapter.output_file
    assert "Model.inp" == tool_adapter.model_file
    assert "15T_2d_mech.rst" == tool_adapter.rst_file
    assert "out_tmp_mech.txt" == tool_adapter.out_temp_file
    assert "-aa_r" == tool_adapter.license_type
    assert "C:\\Program Files\\ANSYS Inc\\v202\\ansys\\bin\\winx64\\ANSYS202.exe" == tool_adapter.executable_path


def test_init_with_json_terminal_ansys():
    # arrange
    config_path = create_resources_file_path('resources/tool_adapters/ansys/terminal_ansys_windows.json')

    # act
    tool_adapter = ToolAdapterFactory.init_with_json(config_path)

    # assert
    assert os.getcwd() == tool_adapter.root_dir
    assert "input\\Mechanic_Plate" == tool_adapter.input_folder_rel_dir
    assert "15T_mech.inp" == tool_adapter.input_file
    assert "vallone.out" == tool_adapter.output_file
    assert "Model.inp" == tool_adapter.model_file
    assert "15T_2d_mech.rst" == tool_adapter.rst_file
    assert "out_tmp_mech.txt" == tool_adapter.out_temp_file
    assert "-aa_r" == tool_adapter.license_type
    assert "C:\\Program Files\\ANSYS Inc\\v202\\ansys\\bin\\winx64\\ANSYS202.exe" == tool_adapter.executable_path


def test_init_with_dict_mapdl_ansys():
    # arrange
    config = {"input_folder_rel_dir": "input\\Mechanic_Plate",
              "input_file": "15T_mech.inp",
              "output_file": "vallone.out",
              "model_file": "Model.inp",
              "rst_file": "15T_2d_mech.rst",
              "single_upload_files": [],
              "reupload_files": []}

    # act
    tool_adapter = ToolAdapterFactory.init_with_dict(config)

    # assert
    assert os.getcwd() == tool_adapter.root_dir
    assert "input\\Mechanic_Plate" == tool_adapter.input_folder_rel_dir
    assert "15T_mech.inp" == tool_adapter.input_file
    assert "vallone.out" == tool_adapter.output_file
    assert "Model.inp" == tool_adapter.model_file
    assert "15T_2d_mech.rst" == tool_adapter.rst_file
    assert [] == tool_adapter.single_upload_files
    assert [] == tool_adapter.reupload_files


def test_init_with_json_mapdl_ansys():
    # arrange
    config_path = create_resources_file_path('resources/tool_adapters/ansys/mapdl_ansys.json')

    # act
    tool_adapter = ToolAdapterFactory.init_with_json(config_path)

    # assert
    assert os.getcwd() == tool_adapter.root_dir
    assert "input\\Mechanic_Plate" == tool_adapter.input_folder_rel_dir
    assert "15T_mech.inp" == tool_adapter.input_file
    assert "vallone.out" == tool_adapter.output_file
    assert "Model.inp" == tool_adapter.model_file
    assert "15T_2d_mech.rst" == tool_adapter.rst_file
    assert [] == tool_adapter.single_upload_files
    assert [] == tool_adapter.reupload_files


def test_init_with_dict_docker_mapdl_ansys():
    # arrange
    config = {"input_folder_rel_dir": "input\\Mechanic_Plate",
              "input_file": "15T_mech.inp",
              "output_file": "vallone.out",
              "model_file": "Model.inp",
              "rst_file": "15T_2d_mech.rst",
              "single_upload_files": [],
              "reupload_files": [],
              "ip": "127.0.0.1",
              "port": 50052}

    # act
    tool_adapter = ToolAdapterFactory.init_with_dict(config)

    # assert
    assert os.getcwd() == tool_adapter.root_dir
    assert "input\\Mechanic_Plate" == tool_adapter.input_folder_rel_dir
    assert "15T_mech.inp" == tool_adapter.input_file
    assert "vallone.out" == tool_adapter.output_file
    assert "Model.inp" == tool_adapter.model_file
    assert "15T_2d_mech.rst" == tool_adapter.rst_file
    assert [] == tool_adapter.single_upload_files
    assert [] == tool_adapter.reupload_files
    assert "127.0.0.1" == tool_adapter.ip
    assert 50052 == tool_adapter.port


def test_init_with_json_docker_mapdl_ansys():
    # arrange
    config_path = create_resources_file_path('resources/tool_adapters/ansys/docker_mapdl_ansys.json')

    # act
    tool_adapter = ToolAdapterFactory.init_with_json(config_path)

    # assert
    assert os.getcwd() == tool_adapter.root_dir
    assert "input\\Mechanic_Plate" == tool_adapter.input_folder_rel_dir
    assert "15T_mech.inp" == tool_adapter.input_file
    assert "vallone.out" == tool_adapter.output_file
    assert "Model.inp" == tool_adapter.model_file
    assert "15T_2d_mech.rst" == tool_adapter.rst_file
    assert [] == tool_adapter.single_upload_files
    assert [] == tool_adapter.reupload_files
    assert "127.0.0.1" == tool_adapter.ip
    assert 50052 == tool_adapter.port



import os
import unittest

import pytest

from magnumapi.tool_adapters.ansys.AnsysToolAdapter import AnsysToolAdapter
from magnumapi.tool_adapters.roxie.RoxieToolAdapter import RoxieToolAdapter
from tests.resource_files import create_resources_file_path

config = {"input_folder_rel_dir": "input%sMechanic_Plate" % os.sep,
          "input_file": "15T_mech.inp",
          "output_file": "vallone.out",
          "model_file": "Model.inp",
          "rst_file": "15T_2d_mech.rst"}


class TestAnsysToolAdapter(unittest.TestCase):

    def test_get_input_path(self):
        # arrange
        ansys = AnsysToolAdapter(**config)

        # act
        input_path = ansys.get_input_path()

        # assert
        input_path_ref = os.path.join(os.getcwd(), 'input%sMechanic_Plate%s15T_mech.inp' % (os.sep, os.sep))
        assert input_path_ref == input_path

    def test_get_output_path(self):
        # arrange
        ansys = AnsysToolAdapter(**config)

        # act
        output_path = ansys.get_output_path()

        # assert
        output_path_ref = os.path.join(os.getcwd(), 'input%sMechanic_Plate%svallone.out' % (os.sep, os.sep))
        assert output_path_ref == output_path

    def test_get_model_path(self):
        # arrange
        ansys = AnsysToolAdapter(**config)

        # act
        model_path = ansys.get_model_path()

        # assert
        model_path_ref = os.path.join(os.getcwd(), 'input%sMechanic_Plate%sModel.inp' % (os.sep, os.sep))
        assert model_path_ref == model_path

    def test_convert_roxie_force_file_to_ansys(self):
        # arrange
        roxie_force_input_path = create_resources_file_path('resources/tool_adapters/ansys/roxie.force2d')
        roxie_force_output_path = create_resources_file_path('resources/tool_adapters/ansys/roxie_edit.force2d')
        ansys_force_output_path = create_resources_file_path('resources/tool_adapters/ansys/forces_edit.vallone')
        field = 16
        target_field = 15

        # act
        # # update ROXIE force file
        RoxieToolAdapter.update_force2d_with_field_scaling(roxie_force_input_path,
                                                           roxie_force_output_path,
                                                           field,
                                                           target_field)

        # # prepare ANSYS force file
        AnsysToolAdapter.convert_roxie_force_file_to_ansys(roxie_force_output_path, ansys_force_output_path)

        # assert
        with open(ansys_force_output_path, 'r') as file:
            ansys_force = file.readlines()

        output_text_ref_path = create_resources_file_path('resources/tool_adapters/ansys/forces_edit_ref.vallone')

        with open(output_text_ref_path, 'r') as file:
            ansys_force_ref = file.readlines()

        for ansys_force_ref_el, ansys_force_el in zip(ansys_force_ref, ansys_force):
            self.assertEqual(ansys_force_ref_el, ansys_force_el)


if __name__ == '__main__':
    unittest.main()




import os
import time
import platform
from subprocess import Popen, PIPE
import xml.etree.ElementTree as ET

import pandas as pd
import numpy as np
import plotly.express as px

from magnumapi.tool_adapters.ToolAdapter import ToolAdapter
import magnumapi.tool_adapters.roxie.RoxieAPI as RoxieAPI
import magnumapi.commons.text_file as text_file

END_OF_ROXIE_CALCULATION = '     *                 END OF THIS ROXIE CALCULATION                 * '


class RoxieToolAdapter(ToolAdapter):
    """ A RoxieToolAdapter class with methods to execute a ROXIE model from command line, read figures of merit table
    and prepare a Lorentz force file

    """

    def __init__(self,
                 input_folder_rel_dir: str,
                 input_file: str,
                 output_file: str,
                 cadata_file: str,
                 xml_output_file: str) -> None:
        """ A constructor of a RoxieToolAdapter instance

        :param input_folder_rel_dir: a relative path to a directory with inputs
        :param input_file_path: a name of an input file, .data
        :param output_file_path: a name of an output file, .output
        :param cadata_file_path: a name of a cadata file, .cadata
        :param xml_output_file_path: a name of an xml output file with plots, .xml
        """
        self.input_folder_rel_dir = input_folder_rel_dir
        self.input_file_path = os.path.join(input_folder_rel_dir, input_file)
        self.output_file_path = os.path.join(input_folder_rel_dir, output_file)
        self.cadata_file_path = os.path.join(input_folder_rel_dir, cadata_file)
        self.cadata_file = cadata_file
        self.xml_output_file_path = os.path.join(input_folder_rel_dir, xml_output_file)
        self.output_lines = []

    def run(self) -> None:
        # if input.output file exists, delete it before running the model
        if os.path.isfile(self.output_file_path):
            os.remove(self.output_file_path)

        # Different execution modes depending on the operating system
        command = self._get_process_command()
        p = Popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        byte_output, _ = p.communicate(b"input data that is passed to subprocess' stdin")

        byte_output_decode = byte_output.decode()
        self.output_lines = byte_output_decode.split('\n')

    def _get_process_command(self) -> list:
        """ Private method returning a command for a process instantiation

        """
        raise NotImplementedError('This method is not implemented for this class')

    def display_output(self) -> None:
        """ Method displaying output in case of ROXIE simulation failure

        """
        if END_OF_ROXIE_CALCULATION not in self.output_lines:
            for output_line in self.output_lines:
                print(output_line)

    @staticmethod
    def convert_figures_of_merit_to_dict(fom_df: pd.DataFrame) -> dict:
        b3 = fom_df[(fom_df.index == 'B') & (fom_df['S1'] == 3)]['OBJECTIVE.1'].values[0]
        b5 = fom_df[(fom_df.index == 'B') & (fom_df['S1'] == 5)]['OBJECTIVE.1'].values[0]
        margmi = fom_df[fom_df.index == 'MARGMI']['OBJECTIVE.1'].values[0]
        bigb = fom_df[fom_df.index == 'BIGB']['OBJECTIVE.1'].values[0]

        return {'b3': b3, 'b5': b5, 'margmi': margmi, 'bigb': bigb}

    def read_figures_of_merit_table(self) -> pd.DataFrame:
        N_OBJECTIVES_KEYWORD = 'NUMBER OF OBJECTIVES AND CONSTRAINTS'
        output_file = open(self.output_file_path, 'r')
        output_lines = output_file.read().split('\n')
        output_file.close()
        idx_start_obj_table, n_objectives = RoxieAPI.find_index_start_and_length_objective_table(output_lines,
                                                                                                 N_OBJECTIVES_KEYWORD)
        objective_table_df = RoxieAPI.extract_objective_table(output_lines, idx_start_obj_table, n_objectives)
        return objective_table_df

    @staticmethod
    def parse_roxie_xml(file_path: str) -> pd.DataFrame:
        """ Static method parsing a ROXIE XML file and returning a dataframe with strand positions and field values

        :param file_path: a path to a file
        :return: output dataframe with strand positions and field values
        """
        mytree = ET.parse(file_path)
        mytree.getroot().get('x')

        x_values = []
        y_values = []

        for type_tag in mytree.getroot().findall('loop/step/loop/step/loop/coilGeom/strands/p'):
            x_value = float(type_tag.get('x'))
            y_value = float(type_tag.get('y'))
            x_values.append(x_value)
            y_values.append(y_value)

        x_strands = []
        y_strands = []
        for i in range(0, len(x_values), 4):
            x_strands.append(np.mean(x_values[i:i + 4]))
            y_strands.append(np.mean(y_values[i:i + 4]))

        if not mytree.getroot().findall('loop/step/loop/step/loop/step/coilData'):
            raise IndexError('The XML file does not contain coilData with results. '
                             'Please enable plotting in the ROXIE file')

        data_str = mytree.getroot().findall('loop/step/loop/step/loop/step/coilData')[0].text \
            .replace(';', '').split('\n')
        data = []
        for data_el in data_str:
            if ',' in data_el:
                data.append(float(data_el.split(',')[1]))

        return pd.DataFrame({'x, [mm]': x_strands, 'y, [mm]': y_strands, '|B| (T)': data})

    @staticmethod
    def plotly_results(strand_data: pd.DataFrame, xlim=(0, 80), ylim=(0, 80)) -> None:
        """ Static method plotting results with plotly package

        :param strand_data: a dataframe with strand positions and field values
        :param xlim: limits in x-direction
        :param ylim: limits in y-direction
        """
        roxie = ['rgb(17,0,181)', 'rgb(11,42,238)', 'rgb(0,104,236)', 'rgb(0,154,235)', 'rgb(0,208,226)',
                 'rgb(0,246,244)', 'rgb(0,255,0)', 'rgb(116,255,0)', 'rgb(191,255,0)', 'rgb(252,255,0)',
                 'rgb(255,199,0)', 'rgb(255,146,0)', 'rgb(255,86,0)', 'rgb(255,0,0)', 'rgb(246,0,0)',
                 'rgb(232, 0, 0)', 'rgb(197, 0, 50)', 'rgb(168, 0, 159)']

        fig = px.scatter(strand_data, x="x, [mm]", y="y, [mm]", color="|B| (T)", hover_data=['|B| (T)'],
                         color_continuous_scale=roxie)
        fig.update_layout(autosize=False,
                          width=750,
                          height=600,
                          xaxis_range=xlim,
                          yaxis_range=ylim,
                          plot_bgcolor='rgba(0,0,0,0)',
                          images=[dict(
                              source="https://i.ibb.co/kcc2mbw/ROXIE.png",
                              xref="paper", yref="paper",
                              x=1.16, y=-0.1,
                              sizex=0.2, sizey=0.2,
                              xanchor="right", yanchor="bottom"
                          )])
        fig.show()

    @staticmethod
    def update_force2d_with_field_scaling(roxie_force_input_path: str,
                                          roxie_force_output_path: str,
                                          field: float,
                                          target_field: float) -> None:
        """ Static method scaling ROXIE force with actual and target field

        :param roxie_force_input_path: input path with ROXIE Lorentz force
        :param roxie_force_output_path: output path with ROXIE Lorentz force
        :param field: actual field in tesla
        :param target_field: target field in tesla
        """
        roxie_force_txt = text_file.readlines(roxie_force_input_path)

        # # convert text input to a list of floats
        scaling = (target_field / field) ** 2
        roxie_force = []
        for roxie_force_txt_el in roxie_force_txt:
            row_float = [float(el) for el in roxie_force_txt_el.replace('\n', '').split(' ') if el != '']
            row_float[2] *= scaling
            row_float[3] *= scaling
            roxie_force.append(row_float)

        with open(roxie_force_output_path, 'w') as file_write:
            for roxie_force_el in roxie_force:
                row_str = [str(roxie_force_el_el) for roxie_force_el_el in roxie_force_el]
                file_write.write(' '.join(row_str) + '\n')

    @staticmethod
    def correct_xml_file(input_xml_path: str, output_xml_path: str) -> None:
        """ Static method for correcting an XML file by removing the first empty line and a first space in the second
        line. The corrected XML file is valid and can be automatically parsed.

        :param input_xml_path: path of an input XML file
        :param output_xml_path: path of an output, corrected XML file
        """
        xml_lines = text_file.read(input_xml_path)

        # get rid of an extra first line and a starting empty space in the second line
        if xml_lines[:2] == '\n ':
            xml_lines = xml_lines[2:]

        text_file.write(output_xml_path, xml_lines)


class TerminalRoxieToolAdapter(RoxieToolAdapter):

    def __init__(self,
                 input_folder_rel_dir: str,
                 input_file: str,
                 output_file: str,
                 cadata_file: str,
                 xml_output_file: str,
                 executable_name: str = 'runroxie'
                 ):
        """ Constructor of a TerminalRoxieToolAdapter object

        :param input_folder_rel_dir: a relative path to input folder w.r.t. notebook location
        :param input_file: a name of an input file
        :param output_file: a name of an output file. It is the input file name with .output extension
        :param cadata_file: a name of a cadata file
        :param xml_output_file: a name of an output XML file with plot information from ROXIE
        :param executable_name: name of ROXIE executable, typically runroxie
        """
        super().__init__(input_folder_rel_dir=input_folder_rel_dir,
                         input_file=input_file,
                         output_file=output_file,
                         cadata_file=cadata_file,
                         xml_output_file=xml_output_file)
        self.executable_name = executable_name

    def _get_process_command(self):
        return [self.executable_name, self.input_file_path]


class DockerTerminalRoxieToolAdapter(TerminalRoxieToolAdapter):

    def __init__(self,
                 input_folder_rel_dir: str,
                 input_file: str,
                 output_file: str,
                 cadata_file: str,
                 xml_output_file: str,
                 executable_name: str = 'runroxie',
                 docker_image_name: str = 'roxie_terminal'
                 ):
        """ Constructor of a DockerTerminalRoxieToolAdapter object

        :param input_folder_rel_dir: a relative path to input folder w.r.t. notebook location
        :param input_file: a name of an input file
        :param output_file: a name of an output file. It is the input file name with .output extension
        :param cadata_file: a name of a cadata file
        :param xml_output_file: a name of an output XML file with plot information from ROXIE
        :param executable_name: a name of ROXIE executable, typically runroxie
        :param docker_image_name: a name of ROXIE Docker terminal image executable, typically roxie_terminal
        """
        super().__init__(input_folder_rel_dir=input_folder_rel_dir,
                         input_file=input_file,
                         output_file=output_file,
                         cadata_file=cadata_file,
                         xml_output_file=xml_output_file,
                         executable_name=executable_name)
        self.docker_image_name = docker_image_name
        self.file_to_unlink = os.path.join(input_folder_rel_dir, 'm4_A.inp')

    def run(self):
        super().run()
        self.wait_until_output_exist(timeout_in_sec=10)
        self.unlink_file(self.file_to_unlink)

    def _get_process_command(self):
        if platform.system() == 'Windows':
            input_file_path_temp = self.input_file_path.replace('\\', '/')
        else:
            input_file_path_temp = self.input_file_path

        return ['docker', 'exec', self.docker_image_name, self.executable_name, input_file_path_temp]

    def wait_until_output_exist(self, timeout_in_sec=10) -> None:
        """ Method waiting until an output file exists. The waiting time is given as an argument. In case of a timeout
         a FileNotFoundError is raised

        :param timeout_in_sec: timeout in seconds
        """
        output_file_path_corrected = self.output_file_path.replace('\\', '/')
        count = 0
        while not os.path.isfile(output_file_path_corrected) and count < timeout_in_sec:
            time.sleep(1)
            count += 1

        if not os.path.isfile(output_file_path_corrected):
            raise FileNotFoundError('The file %s does not exist!' % output_file_path_corrected)

    @staticmethod
    def unlink_file(file_path: str) -> None:
        """ Static method unlinking a file that otherwise causes an error in copying files

        :param file_path: a path to a file to unlink
        """
        if platform.system() == 'Windows':
            file_path = file_path.replace('\\', '/')

        try:
            os.unlink(file_path)
        except:
            print('File does not exist' % file_path)



import ast
from typing import Union


class LayerDefinition:
    """Class for layer definition.
    """

    def __init__(self, no: int, symm: int, typexy: int, blocks: Union[str, list]) -> None:
        """ A constructor of a layer definition instance

        :param no: the row index of a layer definition
        :param symm: multipolar symmetry
        :param typexy: It takes four values:
        - None: The blocks are not duplicated. In 3D one semi-arc of an n-polar coil end is generated.
        - All: All blocks of an n-polar coil are generated. In 3D full coil arcs are generated.
        - One coil: The selected blocks are duplicated only once and placed as to constitute one coil of an n-polar
        cross-section. In 3D a full arc is generated
        - Connection side: the selected blocks are duplicated once per coil in an n-polar cross-section. In 3D only
        semi-arcs are generated. In a connection-side coil end of the asymmetric second half needs to be generated
        separately.
        :param blocks: a list of block indices belonging to a layer
        """
        self.no = no
        self.symm = symm
        self.typexy = typexy
        if isinstance(blocks, str):
            self.blocks = ast.literal_eval(blocks)
        else:
            self.blocks = blocks

    def to_roxie_dict(self):
        return dict(no=self.no,
                    symm=self.symm,
                    typexy=self.typexy,
                    blocks=' '.join([str(block) for block in self.blocks]))


class SlottedLayerDefinition(LayerDefinition):

    def __init__(self, no: int, symm: int, typexy: int, spar_thickness: float, blocks: Union[str, list]) -> None:
        """ A constructor of a layer definition instance

        :param spar_thickness: a thickness of a layer spar in mm
        """
        super().__init__(no=no, symm=symm, typexy=typexy, blocks=blocks)
        self.spar_thickness = spar_thickness



from abc import ABC
import datetime
from copy import deepcopy
from typing import List, Union
from itertools import chain

import pandas as pd

from magnumapi.geometry.GeometryPlot import GeometryPlot
from magnumapi.geometry.blocks.RectangularBlock import RectangularBlock
from magnumapi.geometry.primitives.Line import Line
from magnumapi.geometry.primitives.Area import Area
from magnumapi.geometry.definitions.LayerDefinition import LayerDefinition, SlottedLayerDefinition
from magnumapi.geometry.definitions.RelativeCosThetaBlockDefinition import RelativeCosThetaBlockDefinition
from magnumapi.geometry.definitions.AbsoluteCosThetaBlockDefinition import AbsoluteCosThetaBlockDefinition
from magnumapi.geometry.blocks.Block import Block
from magnumapi.geometry.blocks.CosThetaBlock import RelativeCosThetaBlock, HomogenizedCosThetaBlock, CosThetaBlock, \
    AbsoluteCosThetaBlock

print('Analysis executed on %s' % str(datetime.datetime.now()).split('.')[0])
print('Loaded MagNum API version 0.0.1')


class Geometry(ABC):
    """ Abstract geometry class providing a skeleton for all geometry types (CosTheta, Rectangular, etc.).

    """

    def __init__(self,
                 blocks: List[Union[AbsoluteCosThetaBlock, RectangularBlock,
                                    HomogenizedCosThetaBlock, RelativeCosThetaBlock]],
                 layer_defs: List[Union[LayerDefinition, SlottedLayerDefinition]]
                 ) -> None:
        """ Constructor of a Geometry class

        :param blocks: a list of instances of Block class implementations (e.g., RectangularBlock, CosThetaBlock, etc.)
        :param layer_defs: a list of layer definitions indicating symmetry type,
        and a list of blocks belonging to a layer
        """
        _check_input_consistency(blocks, layer_defs)
        self.blocks = blocks
        self.layer_defs = layer_defs

    def compute_surface(self):
        """ Method computing surface of a geometry as sum of all block surfaces

        :return: a surface in mm^2
        """
        return sum([block.compute_surface() for block in self.blocks])

    def compute_surface_cu(self):
        """ Method computing copper surface of a block as S = ns * S_strand_cu,
        where S_strand_cu = f_cu * pi * d_strand^2 / 4

        :return: a copper surface of a block in mm^2
        """
        return sum([block.compute_surface_cu() for block in self.blocks])

    def compute_surface_nocu(self):
        """ Method computing copper surface of a block as S = ns * S_strand_nocu,
        where S_strand_nocu = f_nocu * pi * d_strand^2 / 4

        :return: a non-copper surface of a block in mm^2
        """
        return sum([block.compute_surface_nocu() for block in self.blocks])

    def build_blocks(self) -> None:
        """ Method building all blocks for a given geometry definition

        """
        for block in self.blocks:
            block.build_block()

    def to_block_df(self) -> pd.DataFrame:
        """ Method concatenates row definition of each block into a ROXIE-compatible dataframe

        :return: a concatenated dataframe with ROXIE block definition for a geometry instance
        """
        return pd.concat([block.to_block_df() for block in self.blocks], axis=0).reset_index(drop=True)

    def to_layer_df(self) -> pd.DataFrame:
        return pd.DataFrame([layer_def.to_roxie_dict() for layer_def in self.layer_defs])

    def to_dict(self) -> dict:
        """ Method returning an absolute geometry definition (block definitions and layer definitions) as a dictionary.

        :return: a dictionary with two keys: block_defs and layer_defs
        """
        block_defs = []
        for layer_def in self.layer_defs:
            for block_index in layer_def.blocks:
                index_in_blocks = self.get_index_in_blocks_for_layer_block_index(block_index)
                block_defs.append(self.blocks[index_in_blocks].to_abs_dict())

        layer_defs = [layer_def.__dict__ for layer_def in self.layer_defs]

        return self.init_geometry_dict(block_defs, layer_defs)

    def init_geometry_dict(self, block_defs, layer_defs):
        return {'block_defs': block_defs, 'layer_defs': layer_defs}

    def to_df(self) -> pd.DataFrame:
        """ Method concatenating dataframes with area coordinates for each block

        :return: a concatenated dataframe with area coordinates
        """
        return pd.concat([block.to_df() for block in self.blocks], axis=0).reset_index(drop=True)

    def plot_blocks(self, figsize=(10, 10), is_grid=True, xlim=(0, 80), ylim=(0, 80)) -> None:
        GeometryPlot.plot_blocks(self, figsize, is_grid, xlim, ylim)

    def plotly_blocks(self, figsize=(750, 750), xlim=(0, 80), ylim=(0, 80)) -> None:
        GeometryPlot.plotly_geometry_blocks(self, figsize, xlim, ylim)

    def plot_bare_blocks(self, figsize=(15, 15), is_grid=True, xlim=(0, 80), ylim=(0, 80)) -> None:
        GeometryPlot.plot_bare_blocks(self, figsize, is_grid, xlim, ylim)

    def get_bare_areas_for_blocks(self) -> List[List[Area]]:
        """ Method returning a list of list of bare (uninsulated) areas constructed from iterating over all blocks in a
        geometry and all areas in a block.

        :return: a list of list of bare (uninsulated) areas
        """
        return [block.get_bare_areas() for block in self.blocks]

    def empty_block_areas(self):
        for block in self.blocks:
            block.empty_areas()

    def to_rel_geometry(self) -> "RelativeCosThetaGeometry":
        geometry = deepcopy(self)
        # check whether areas were initialize before geometry casting
        is_any_area_initialized = any([area for block in self.blocks for area in block.areas])

        if not all([isinstance(block, CosThetaBlock) for block in self.blocks]):
            raise TypeError('Only a geometry composed of CosThetaBlocks can be converted to a relative geometry.')

        geometry.build_blocks()

        blocks = []
        for layer_def in geometry.layer_defs:
            for index, block_index in enumerate(layer_def.blocks):
                index_in_blocks = geometry.get_index_in_blocks_for_layer_block_index(block_index)
                block = geometry.blocks[index_in_blocks]
                alpha_ref, phi_ref = retrieve_alpha_and_phi_ref(geometry, layer_def, index)
                block_def = RelativeCosThetaBlockDefinition(**block.to_rel_dict(alpha_ref=alpha_ref, phi_ref=phi_ref))
                blocks.append(RelativeCosThetaBlock(cable_def=block.cable_def,
                                                    insul_def=block.insul_def,
                                                    strand_def=block.strand_def,
                                                    conductor_def=block.conductor_def,
                                                    block_def=block_def))

        geometry_rel = self.init_geometry_instance(blocks, geometry)
        # Return to the external context of the geometry before entering
        Geometry.recover_context(is_any_area_initialized, geometry_rel)
        return geometry_rel

    def init_geometry_instance(self, blocks, geometry):
        return RelativeCosThetaGeometry(blocks=blocks, layer_defs=geometry.layer_defs)

    def homogenize(self) -> "HomogenizedCosThetaGeometry":
        """ Class method creating a homogenized cos-theta geometry from a cos-theta geometry.
        The method assumes that all blocks are CosThetaBlock type

        :param geometry: an input cos-theta geometry
        :return: a HomogenizedCosThetaGeometry instance
        """
        if not all([isinstance(block, CosThetaBlock) for block in self.blocks]):
            raise TypeError('Only a geometry composed of CosThetaBlocks can be converted to a homogenized geometry.')

        blocks = []
        for layer_index, layer_def in enumerate(self.layer_defs):
            for block_index in layer_def.blocks:
                index_in_blocks = self.get_index_in_blocks_for_layer_block_index(block_index)
                block = self.blocks[index_in_blocks]
                homo_cos_theta_block = block.homogenize()
                blocks.append(homo_cos_theta_block)

        return self.init_homogenized_geometry(blocks, self.layer_defs)

    def init_homogenized_geometry(self, blocks, layer_defs):
        return HomogenizedCosThetaGeometry(blocks, layer_defs)

    def get_index_in_blocks_for_layer_block_index(self, block_index: int) -> int:
        # The indexing of blocks does not have to be incremented from block to block
        # - also in a layer the block indexing is not given
        for index, block in enumerate(self.blocks):
            if block.block_def.no == block_index:
                return index

        raise IndexError('Block index from layer %d does not appear in block definitions' % block_index)

    @staticmethod
    def recover_context(is_any_area_initialized, geometry_rel):
        if is_any_area_initialized:
            geometry_rel.build_blocks()
        else:
            geometry_rel.empty_block_areas()

    def to_abs_geometry(self) -> "Geometry":
        return deepcopy(self)

    def get_number_of_layers(self) -> int:
        """ Method returning the number of layers in a cos-theta coil

        :return: number of layers
        """
        return len(self.layer_defs)

    def get_number_of_blocks_per_layer(self) -> List[int]:
        """ Method returning the number of blocks per layer in a cos-theta coil

        :return: list with number of blocks per layer
        """
        return [len(layer_def.blocks) for layer_def in self.layer_defs]


def _check_input_consistency(blocks: List[Block], layer_defs: List[LayerDefinition]) -> None:
    block_nos = [block.block_def.no for block in blocks]
    # are there any duplications in block numbers
    if len(block_nos) != len(set(block_nos)):
        raise AttributeError('The block numbering ({}) contains duplications!'.format(block_nos))

    # are there any duplications in layer numbers
    layer_nos = [layer_def.no for layer_def in layer_defs]
    if len(layer_nos) != len(set(layer_nos)):
        raise AttributeError('The layer numbering ({}) contains duplications!'.format(layer_nos))

    # are there any duplications in layer block numbers
    layer_blocks = list(chain.from_iterable([layer_def.blocks for layer_def in layer_defs]))
    if len(layer_blocks) != len(set(layer_blocks)):
        raise AttributeError('The layer numbering of blocks ({}) contains duplications!'.format(layer_blocks))

    # are the block nos matching the layer block numbers
    if set(block_nos) != set(layer_blocks):
        raise AttributeError('The numbering in block {} and layer {} do not match!'.format(block_nos, layer_blocks))


def retrieve_alpha_and_phi_ref(geometry, layer_def, index):
    if index == 0:
        return 0.0, 0.0
    else:
        index_prev_in_blocks = geometry.get_index_in_blocks_for_layer_block_index(layer_def.blocks[index - 1])
        block_prev = geometry.blocks[index_prev_in_blocks]
        area_prev = block_prev.areas[-1]
        radius = block_prev.get_radius()
        phi_ref = Line.calculate_positioning_angle(area_prev.get_line(2), radius)
        alpha_ref = Line.calculate_relative_alpha_angle(area_prev.get_line(2))
        return alpha_ref, phi_ref


class RelativeCosThetaGeometry(Geometry):
    """RelativeCosThetaGeometry class for relative cos-theta geometry. Needed to implement the relative creation of
    cos-theta blocks.

    """

    def __init__(self, blocks: List[RelativeCosThetaBlock], layer_defs: List[LayerDefinition]) -> None:
        """Constructor of RelativeCosThetaGeometry class

        :param blocks: list of RelativeCosThetaBlock definitions
        :param layer_defs: a list of layer definitions
        """
        super().__init__(blocks, layer_defs)
        self.blocks = blocks  # Superfluous assignment to fix attribute warnings of mypy

    def build_blocks(self):
        for layer_index, layer_def in enumerate(self.layer_defs):
            for index, block_index in enumerate(layer_def.blocks):
                index_in_blocks = self.get_index_in_blocks_for_layer_block_index(block_index)
                block = self.blocks[index_in_blocks]
                alpha_ref, phi_ref = retrieve_alpha_and_phi_ref(self, layer_def, index)
                block.build_block(phi_ref=phi_ref, alpha_ref=alpha_ref)

    def to_abs_geometry(self) -> "Geometry":
        geometry = deepcopy(self)
        # check whether areas were initialize before geometry casting
        is_any_area_initialized = any([area for block in self.blocks for area in block.areas])

        geometry.build_blocks()

        blocks = []
        for block in geometry.blocks:
            abs_block_def = AbsoluteCosThetaBlockDefinition(**block.to_abs_dict())
            abs_block = AbsoluteCosThetaBlock(cable_def=block.cable_def,
                                              insul_def=block.insul_def,
                                              strand_def=block.strand_def,
                                              conductor_def=block.conductor_def,
                                              block_def=abs_block_def)
            blocks.append(abs_block)

        geometry_abs = self.init_geometry_instance(blocks, geometry.layer_defs)
        # Return to the external context of the geometry before entering
        Geometry.recover_context(is_any_area_initialized, geometry_abs)
        return geometry_abs

    def init_geometry_instance(self, blocks, layer_defs):
        return Geometry(blocks=blocks, layer_defs=self.layer_defs)

    def to_rel_geometry(self) -> "RelativeCosThetaGeometry":
        return deepcopy(self)

    def to_dict(self) -> dict:
        """ Method returning a relative geometry definition (block definitions and layer definitions) as a dictionary.

        :return: a dictionary with two keys: block_defs and layer_defs
        """
        block_defs = []
        for layer_def in self.layer_defs:
            for block_index in layer_def.blocks:
                index_in_blocks = self.get_index_in_blocks_for_layer_block_index(block_index)
                block_defs.append(self.blocks[index_in_blocks].to_rel_dict())

        layer_defs = [layer_def.__dict__ for layer_def in self.layer_defs]

        return self.init_geometry_dict(block_defs, layer_defs)

    def init_geometry_dict(self, block_defs, layer_defs):
        return {'block_defs': block_defs, 'layer_defs': layer_defs}


class HomogenizedCosThetaGeometry(Geometry):
    """HomogenizedCosThetaGeometry class for homogenized cos-theta geometry. Creates a homogenized geometry from both
    relative and absolute cos-theta geometry definition. Used for creation of ANSYS models.

    """

    def __init__(self, blocks: List[HomogenizedCosThetaBlock], layer_defs: List[LayerDefinition]) -> None:
        """Constructor of HomogenizedCosThetaGeometry class

        :param blocks: list of HomogenizedCosThetaBlock blocks
        :param layer_defs: a list of layer definitions
        """
        super().__init__(blocks, layer_defs)
        self.blocks = blocks  # Superfluous assignment to fix attribute warnings of mypy

    def to_block_df(self):
        raise NotImplementedError('This method is not implemented for this class')

    def plotly_blocks(self, figsize=(750, 750), xlim=(0, 80), ylim=(0, 80)):
        GeometryPlot.plotly_homogenized_geometry_blocks(self, figsize, xlim, ylim)
